######################################################################
#  Do not modify, these classes are autogenerated by gen_classes.py  #
######################################################################

import dateutil.parser


class BaseObject(object):
    def to_dict(self):
        copy_dict = self.__dict__.copy()
        for key in list(copy_dict.keys()):
            if copy_dict[key] is None or key == 'api':
                del copy_dict[key]
                continue

            if key.startswith('_'):
                copy_dict[key[1:]] = copy_dict[key]
                del copy_dict[key]
        return copy_dict

    def __repr__(self):
        if hasattr(self, 'id'):
            return "[%s(id=%s)]" % (self.__class__.__name__, self.id)
        elif hasattr(self, 'token'):
            return "[%s(token='%s')]" % (self.__class__.__name__, self.token)
        else:
            return "[%s()]" % self.__class__.__name__


class NotificationEvent(BaseObject):
    def __init__(self, api=None, body=None, id=None, subject=None, type=None, **kwargs):

        self.api = api

        self._recipients = None

        self._via = None
        self.body = body
        self.id = id
        self.subject = subject
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def recipients(self):

        if self.api and self._recipients:
            return self.api._get_users(self._recipients)

    @recipients.setter
    def recipients(self, recipients):
        if recipients:
            self._recipients = recipients

    @property
    def via(self):

        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via


class Forum(BaseObject):
    def __init__(self, api=None, access=None, category_id=None, created_at=None, description=None, forum_type=None,
                 id=None, locale_id=None, locked=None, name=None, organization_id=None, position=None, tags=None,
                 updated_at=None, url=None, **kwargs):

        self.api = api
        self.access = access
        self.category_id = category_id
        self.created_at = created_at
        self.description = description
        self.forum_type = forum_type
        self.id = id
        self.locale_id = locale_id
        self.locked = locked
        self.name = name
        self.organization_id = organization_id
        self.position = position
        self.tags = tags
        self.updated_at = updated_at
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def category(self):

        if self.api and self.category_id:
            return self.api._get_category(self.category_id)

    @category.setter
    def category(self, category):
        if category:
            self.category_id = category.id
            self._category = category

    @property
    def created(self):

        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def organization(self):

        if self.api and self.organization_id:
            return self.api._get_organization(self.organization_id)

    @organization.setter
    def organization(self, organization):
        if organization:
            self.organization_id = organization.id
            self._organization = organization

    @property
    def updated(self):

        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Tag(BaseObject):
    def __init__(self, api=None, count=None, name=None, **kwargs):
        self.api = api
        self.count = count
        self.name = name

        for key, value in kwargs.items():
            setattr(self, key, value)


class TicketField(BaseObject):
    def __init__(self, api=None, active=None, collapsed_for_agents=None, created_at=None, description=None,
                 editable_in_portal=None, id=None, position=None, raw_description=None, raw_title=None,
                 raw_title_in_portal=None, regexp_for_validation=None, required=None, required_in_portal=None, tag=None,
                 title=None, title_in_portal=None, type=None, updated_at=None, url=None, visible_in_portal=None,
                 **kwargs):

        self.api = api

        #:| Comment: Whether this field is available
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.active = active

        #:| Comment: If this field should be shown to agents by default or be hidden alongside infrequently used fields. Classic interface only
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.collapsed_for_agents = collapsed_for_agents

        #:| Comment: The time the ticket field was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: The description of the purpose of this ticket field, shown to users
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.description = description

        #:| Comment: Whether this field is editable by end users
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.editable_in_portal = editable_in_portal

        #:| Comment: Automatically assigned upon creation
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: A relative position for the ticket fields, determines the order of ticket fields on a ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.position = position

        #:| Comment: The dynamic content placeholder, if present, or the "description" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_description = raw_description

        #:| Comment: The dynamic content placeholder, if present, or the "title" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_title = raw_title

        #:| Comment: The dynamic content placeholder, if present, or the "title_in_portal" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_title_in_portal = raw_title_in_portal

        #:| Comment: Regular expression field only. The validation pattern for a field value to be deemed valid.
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.regexp_for_validation = regexp_for_validation

        #:| Comment: If it's required for this field to have a value when updated by agents
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.required = required

        #:| Comment: If it's required for this field to have a value when updated by end users
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.required_in_portal = required_in_portal

        #:| Comment: A tag value to set for checkbox fields when checked
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.tag = tag

        #:| Comment: The title of the ticket field
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.title = title

        #:| Comment: The title of the ticket field when shown to end users
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.title_in_portal = title_in_portal

        #:| Comment: The type of the ticket field: "checkbox", "date", "decimal", "integer", "regexp", "tagger", "text", or "textarea"
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.type = type

        #:| Comment: The time of the last update of the ticket field
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The URL for this resource
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        #:| Comment: Whether this field is available to end users
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.visible_in_portal = visible_in_portal

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: The time the ticket field was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the ticket field
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class TweetEvent(BaseObject):
    def __init__(self, api=None, body=None, direct_message=None, id=None, type=None, **kwargs):

        self.api = api

        self._recipients = None
        self.body = body
        self.direct_message = direct_message
        self.id = id
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def recipients(self):

        if self.api and self._recipients:
            return self.api._get_users(self._recipients)

    @recipients.setter
    def recipients(self, recipients):
        if recipients:
            self._recipients = recipients


class TicketMetric(BaseObject):
    def __init__(self, api=None, assigned_at=None, assignee_stations=None, assignee_updated_at=None, created_at=None,
                 group_stations=None, id=None, initially_assigned_at=None, latest_comment_added_at=None, reopens=None,
                 replies=None, requester_updated_at=None, solved_at=None, status_updated_at=None, ticket_id=None,
                 updated_at=None, **kwargs):

        self.api = api

        #:| Comment: Number of minutes the agent spent waiting inside and out of business hours
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._agent_wait_time_in_minutes = None

        #:| Comment: Number of minutes to the first resolution time inside and out of business hours
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._first_resolution_time_in_minutes = None

        #:| Comment: Number of minutes to the full resolution inside and out of business hours
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._full_resolution_time_in_minutes = None

        self._on_hold_time_in_minutes = None

        #:| Comment: Number of minutes to the first reply inside and out of business hours
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._reply_time_in_minutes = None

        #:| Comment: Number of minutes the requester spent waiting inside and out of business hours
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._requester_wait_time_in_minutes = None

        #:| Comment: When the ticket was last assigned
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.assigned_at = assigned_at

        #:| Comment: Number of assignees this ticket had
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.assignee_stations = assignee_stations

        #:| Comment: When the assignee last updated the ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.assignee_updated_at = assignee_updated_at

        #:| Comment: When this record was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Number of groups this ticket passed through
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.group_stations = group_stations

        #:| Comment: Automatically assigned
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: When the ticket was initially assigned
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.initially_assigned_at = initially_assigned_at

        #:| Comment: When the latest comment was added
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.latest_comment_added_at = latest_comment_added_at

        #:| Comment: Total number of times the ticket was reopened
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.reopens = reopens

        #:| Comment: Total number of times ticket was replied to
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.replies = replies

        #:| Comment: When the requester last updated the ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.requester_updated_at = requester_updated_at

        #:| Comment: When the ticket was solved
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.solved_at = solved_at

        #:| Comment: When the status was last updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.status_updated_at = status_updated_at

        #:| Comment: Id of the associated ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.ticket_id = ticket_id

        #:| Comment: When this record last got updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def agent_wait_time_in_minutes(self):
        """
        |  Comment: Number of minutes the agent spent waiting inside and out of business hours
        """
        if self.api and self._agent_wait_time_in_minutes:
            return self.api._get_ticket_metric_item(self._agent_wait_time_in_minutes)

    @agent_wait_time_in_minutes.setter
    def agent_wait_time_in_minutes(self, agent_wait_time_in_minutes):
        if agent_wait_time_in_minutes:
            self._agent_wait_time_in_minutes = agent_wait_time_in_minutes

    @property
    def first_resolution_time_in_minutes(self):
        """
        |  Comment: Number of minutes to the first resolution time inside and out of business hours
        """
        if self.api and self._first_resolution_time_in_minutes:
            return self.api._get_ticket_metric_item(self._first_resolution_time_in_minutes)

    @first_resolution_time_in_minutes.setter
    def first_resolution_time_in_minutes(self, first_resolution_time_in_minutes):
        if first_resolution_time_in_minutes:
            self._first_resolution_time_in_minutes = first_resolution_time_in_minutes

    @property
    def full_resolution_time_in_minutes(self):
        """
        |  Comment: Number of minutes to the full resolution inside and out of business hours
        """
        if self.api and self._full_resolution_time_in_minutes:
            return self.api._get_ticket_metric_item(self._full_resolution_time_in_minutes)

    @full_resolution_time_in_minutes.setter
    def full_resolution_time_in_minutes(self, full_resolution_time_in_minutes):
        if full_resolution_time_in_minutes:
            self._full_resolution_time_in_minutes = full_resolution_time_in_minutes

    @property
    def on_hold_time_in_minutes(self):

        if self.api and self._on_hold_time_in_minutes:
            return self.api._get_ticket_metric_item(self._on_hold_time_in_minutes)

    @on_hold_time_in_minutes.setter
    def on_hold_time_in_minutes(self, on_hold_time_in_minutes):
        if on_hold_time_in_minutes:
            self._on_hold_time_in_minutes = on_hold_time_in_minutes

    @property
    def reply_time_in_minutes(self):
        """
        |  Comment: Number of minutes to the first reply inside and out of business hours
        """
        if self.api and self._reply_time_in_minutes:
            return self.api._get_ticket_metric_item(self._reply_time_in_minutes)

    @reply_time_in_minutes.setter
    def reply_time_in_minutes(self, reply_time_in_minutes):
        if reply_time_in_minutes:
            self._reply_time_in_minutes = reply_time_in_minutes

    @property
    def requester_wait_time_in_minutes(self):
        """
        |  Comment: Number of minutes the requester spent waiting inside and out of business hours
        """
        if self.api and self._requester_wait_time_in_minutes:
            return self.api._get_ticket_metric_item(self._requester_wait_time_in_minutes)

    @requester_wait_time_in_minutes.setter
    def requester_wait_time_in_minutes(self, requester_wait_time_in_minutes):
        if requester_wait_time_in_minutes:
            self._requester_wait_time_in_minutes = requester_wait_time_in_minutes

    @property
    def assigned(self):
        """
        |  Comment: When the ticket was last assigned
        """
        if self.assigned_at:
            return dateutil.parser.parse(self.assigned_at)

    @assigned.setter
    def assigned(self, assigned):
        if assigned:
            self.assigned_at = assigned

    @property
    def assignee_updated(self):
        """
        |  Comment: When the assignee last updated the ticket
        """
        if self.assignee_updated_at:
            return dateutil.parser.parse(self.assignee_updated_at)

    @assignee_updated.setter
    def assignee_updated(self, assignee_updated):
        if assignee_updated:
            self.assignee_updated_at = assignee_updated

    @property
    def created(self):
        """
        |  Comment: When this record was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def initially_assigned(self):
        """
        |  Comment: When the ticket was initially assigned
        """
        if self.initially_assigned_at:
            return dateutil.parser.parse(self.initially_assigned_at)

    @initially_assigned.setter
    def initially_assigned(self, initially_assigned):
        if initially_assigned:
            self.initially_assigned_at = initially_assigned

    @property
    def latest_comment_added(self):
        """
        |  Comment: When the latest comment was added
        """
        if self.latest_comment_added_at:
            return dateutil.parser.parse(self.latest_comment_added_at)

    @latest_comment_added.setter
    def latest_comment_added(self, latest_comment_added):
        if latest_comment_added:
            self.latest_comment_added_at = latest_comment_added

    @property
    def requester_updated(self):
        """
        |  Comment: When the requester last updated the ticket
        """
        if self.requester_updated_at:
            return dateutil.parser.parse(self.requester_updated_at)

    @requester_updated.setter
    def requester_updated(self, requester_updated):
        if requester_updated:
            self.requester_updated_at = requester_updated

    @property
    def solved(self):
        """
        |  Comment: When the ticket was solved
        """
        if self.solved_at:
            return dateutil.parser.parse(self.solved_at)

    @solved.setter
    def solved(self, solved):
        if solved:
            self.solved_at = solved

    @property
    def status_updated(self):
        """
        |  Comment: When the status was last updated
        """
        if self.status_updated_at:
            return dateutil.parser.parse(self.status_updated_at)

    @status_updated.setter
    def status_updated(self, status_updated):
        if status_updated:
            self.status_updated_at = status_updated

    @property
    def ticket(self):
        """
        |  Comment: Id of the associated ticket
        """
        if self.api and self.ticket_id:
            return self.api._get_ticket(self.ticket_id)

    @ticket.setter
    def ticket(self, ticket):
        if ticket:
            self.ticket_id = ticket.id
            self._ticket = ticket

    @property
    def updated(self):
        """
        |  Comment: When this record last got updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class System(BaseObject):
    def __init__(self, api=None, client=None, ip_address=None, latitude=None, location=None, longitude=None, **kwargs):
        self.api = api
        self.client = client
        self.ip_address = ip_address
        self.latitude = latitude
        self.location = location
        self.longitude = longitude

        for key, value in kwargs.items():
            setattr(self, key, value)


class OrganizationField(BaseObject):
    def __init__(self, api=None, active=None, created_at=None, description=None, id=None, key=None, position=None,
                 raw_description=None, raw_title=None, regexp_for_validation=None, title=None, type=None,
                 updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: If true, this field is available for use
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.active = active

        #:| Comment: The time the ticket field was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: User-defined description of this field's purpose
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.description = description

        #:| Comment: Automatically assigned upon creation
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: A unique key that identifies this custom field. This is used for updating the field and referencing in placeholders.
        #:| Mandatory: on create
        #:| Read-only: no
        #:| Type: string
        self.key = key

        #:| Comment: Ordering of the field relative to other fields
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.position = position

        #:| Comment: The dynamic content placeholder, if present, or the "description" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_description = raw_description

        #:| Comment: The dynamic content placeholder, if present, or the "title" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_title = raw_title

        #:| Comment: Regular expression field only. The validation pattern for a field value to be deemed valid.
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.regexp_for_validation = regexp_for_validation

        #:| Comment: The title of the custom field
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.title = title

        #:| Comment: Type of the custom field: "checkbox", "date", "decimal", "dropdown", "integer", "regexp", "text", or "textarea"
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.type = type

        #:| Comment: The time of the last update of the ticket field
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The URL for this resource
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: The time the ticket field was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the ticket field
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class UserRelated(BaseObject):
    def __init__(self, api=None, assigned_tickets=None, ccd_tickets=None, entry_subscriptions=None,
                 forum_subscriptions=None, organization_subscriptions=None, requested_tickets=None, subscriptions=None,
                 topic_comments=None, topics=None, votes=None, **kwargs):
        self.api = api
        self.assigned_tickets = assigned_tickets
        self.ccd_tickets = ccd_tickets
        self.entry_subscriptions = entry_subscriptions
        self.forum_subscriptions = forum_subscriptions
        self.organization_subscriptions = organization_subscriptions
        self.requested_tickets = requested_tickets
        self.subscriptions = subscriptions
        self.topic_comments = topic_comments
        self.topics = topics
        self.votes = votes

        for key, value in kwargs.items():
            setattr(self, key, value)


class TicketEvent(BaseObject):
    def __init__(self, api=None, id=None, ticket_id=None, timestamp=None, updater_id=None, via=None, **kwargs):

        self.api = api

        self._child_events = None
        self.id = id
        self.ticket_id = ticket_id
        self.timestamp = timestamp
        self.updater_id = updater_id
        self.via = via

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def child_events(self):

        if self.api and self._child_events:
            return self.api._get_child_events(self._child_events)

    @child_events.setter
    def child_events(self, child_events):
        if child_events:
            self._child_events = child_events

    @property
    def ticket(self):

        if self.api and self.ticket_id:
            return self.api._get_ticket(self.ticket_id)

    @ticket.setter
    def ticket(self, ticket):
        if ticket:
            self.ticket_id = ticket.id
            self._ticket = ticket

    @property
    def updater(self):

        if self.api and self.updater_id:
            return self.api._get_user(self.updater_id)

    @updater.setter
    def updater(self, updater):
        if updater:
            self.updater_id = updater.id
            self._updater = updater


class VoiceCommentEvent(BaseObject):
    def __init__(self, api=None, author_id=None, body=None, formatted_from=None, formatted_to=None, html_body=None,
                 id=None, public=None, transcription_visible=None, trusted=None, type=None, **kwargs):

        self.api = api

        self._attachments = None

        self._data = None
        self.author_id = author_id
        self.body = body
        self.formatted_from = formatted_from
        self.formatted_to = formatted_to
        self.html_body = html_body
        self.id = id
        self.public = public
        self.transcription_visible = transcription_visible
        self.trusted = trusted
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def attachments(self):

        if self.api and self._attachments:
            return self.api._get_attachments(self._attachments)

    @attachments.setter
    def attachments(self, attachments):
        if attachments:
            self._attachments = attachments

    @property
    def data(self):

        if self.api and self._data:
            return self.api._get_data(self._data)

    @data.setter
    def data(self, data):
        if data:
            self._data = data

    @property
    def author(self):

        if self.api and self.author_id:
            return self.api._get_user(self.author_id)

    @author.setter
    def author(self, author):
        if author:
            self.author_id = author.id
            self._author = author


class Request(BaseObject):
    def __init__(self, api=None, assignee_id=None, can_be_solved_by_me=None, collaborator_ids=None, created_at=None,
                 description=None, due_at=None, id=None, organization_id=None, priority=None, requester_id=None,
                 status=None, subject=None, type=None, updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: The fields and entries for this request
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: :class:`Array`

        self._custom_fields = None

        self._fields = None

        #:| Comment: This object explains how the request was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: :class:`Via`

        self._via = None

        #:| Comment: The id of the assignee if the field is visible to end users
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.assignee_id = assignee_id

        #:| Comment: If true, end user can mark request as solved.
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.can_be_solved_by_me = can_be_solved_by_me

        #:| Comment: Who are currently CC'ed on the ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: array
        self.collaborator_ids = collaborator_ids

        #:| Comment: When this record was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: The first comment on the request
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: string
        self.description = description

        #:| Comment: When the task is due (only applies if the request is of type "task")
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: date
        self.due_at = due_at

        #:| Comment: Automatically assigned when creating requests
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The organization of the requester
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.organization_id = organization_id

        #:| Comment: The priority of the request, "low", "normal", "high", "urgent"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.priority = priority

        #:| Comment: The id of the requester
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.requester_id = requester_id

        #:| Comment: The state of the request, "new", "open", "pending", "hold", "solved", "closed"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.status = status

        #:| Comment: The value of the subject field for this request if the subject field is visible to end users; a truncated version of the description otherwise
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.subject = subject

        #:| Comment: The type of the request, "question", "incident", "problem", "task"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.type = type

        #:| Comment: When this record last got updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this request
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def custom_fields(self):
        """
        |  Comment: The fields and entries for this request
        """
        if self.api and self._custom_fields:
            return self.api._get_custom_fields(self._custom_fields)

    @custom_fields.setter
    def custom_fields(self, custom_fields):
        if custom_fields:
            self._custom_fields = custom_fields

    @property
    def fields(self):

        if self.api and self._fields:
            return self.api._get_fields(self._fields)

    @fields.setter
    def fields(self, fields):
        if fields:
            self._fields = fields

    @property
    def via(self):
        """
        |  Comment: This object explains how the request was created
        """
        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via

    @property
    def assignee(self):
        """
        |  Comment: The id of the assignee if the field is visible to end users
        """
        if self.api and self.assignee_id:
            return self.api._get_user(self.assignee_id)

    @assignee.setter
    def assignee(self, assignee):
        if assignee:
            self.assignee_id = assignee.id
            self._assignee = assignee

    @property
    def collaborators(self):
        """
        |  Comment: Who are currently CC'ed on the ticket
        """
        if self.api and self.collaborator_ids:
            return self.api._get_users(self.collaborator_ids)

    @collaborators.setter
    def collaborators(self, collaborators):
        if collaborators:
            self.collaborator_ids = [o.id for o in collaborators]
            self._collaborators = collaborators

    @property
    def created(self):
        """
        |  Comment: When this record was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def due(self):
        """
        |  Comment: When the task is due (only applies if the request is of type "task")
        """
        if self.due_at:
            return dateutil.parser.parse(self.due_at)

    @due.setter
    def due(self, due):
        if due:
            self.due_at = due

    @property
    def organization(self):
        """
        |  Comment: The organization of the requester
        """
        if self.api and self.organization_id:
            return self.api._get_organization(self.organization_id)

    @organization.setter
    def organization(self, organization):
        if organization:
            self.organization_id = organization.id
            self._organization = organization

    @property
    def requester(self):
        """
        |  Comment: The id of the requester
        """
        if self.api and self.requester_id:
            return self.api._get_user(self.requester_id)

    @requester.setter
    def requester(self, requester):
        if requester:
            self.requester_id = requester.id
            self._requester = requester

    @property
    def updated(self):
        """
        |  Comment: When this record last got updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class UserField(BaseObject):
    def __init__(self, api=None, active=None, created_at=None, description=None, id=None, key=None, position=None,
                 raw_description=None, raw_title=None, regexp_for_validation=None, title=None, type=None,
                 updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: If true, this field is available for use
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.active = active

        #:| Comment: The time the ticket field was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: User-defined description of this field's purpose
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.description = description

        #:| Comment: Automatically assigned upon creation
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: A unique key that identifies this custom field. This is used for updating the field and referencing in placeholders.
        #:| Mandatory: on create
        #:| Read-only: no
        #:| Type: string
        self.key = key

        #:| Comment: Ordering of the field relative to other fields
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.position = position

        #:| Comment: The dynamic content placeholder, if present, or the "description" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_description = raw_description

        #:| Comment: The dynamic content placeholder, if present, or the "title" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_title = raw_title

        #:| Comment: Regular expression field only. The validation pattern for a field value to be deemed valid.
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.regexp_for_validation = regexp_for_validation

        #:| Comment: The title of the custom field
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.title = title

        #:| Comment: Type of the custom field: "checkbox", "date", "decimal", "dropdown", "integer", "regexp", "text", or "textarea"
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.type = type

        #:| Comment: The time of the last update of the ticket field
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The URL for this resource
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: The time the ticket field was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the ticket field
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Metadata(BaseObject):
    def __init__(self, api=None, **kwargs):

        self.api = api

        self._custom = None

        self._system = None

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def custom(self):

        if self.api and self._custom:
            return self.api._get_custom(self._custom)

    @custom.setter
    def custom(self, custom):
        if custom:
            self._custom = custom

    @property
    def system(self):

        if self.api and self._system:
            return self.api._get_system(self._system)

    @system.setter
    def system(self, system):
        if system:
            self._system = system


class Topic(BaseObject):
    def __init__(self, api=None, body=None, created_at=None, forum_id=None, id=None, locked=None, pinned=None,
                 position=None, submitter_id=None, tags=None, title=None, topic_type=None, updated_at=None,
                 updater_id=None, url=None, **kwargs):

        self.api = api

        self._search_phrases = None
        self.body = body
        self.created_at = created_at
        self.forum_id = forum_id
        self.id = id
        self.locked = locked
        self.pinned = pinned
        self.position = position
        self.submitter_id = submitter_id
        self.tags = tags
        self.title = title
        self.topic_type = topic_type
        self.updated_at = updated_at
        self.updater_id = updater_id
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def search_phrases(self):

        if self.api and self._search_phrases:
            return self.api._get_search_phrases(self._search_phrases)

    @search_phrases.setter
    def search_phrases(self, search_phrases):
        if search_phrases:
            self._search_phrases = search_phrases

    @property
    def created(self):

        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def forum(self):

        if self.api and self.forum_id:
            return self.api._get_forum(self.forum_id)

    @forum.setter
    def forum(self, forum):
        if forum:
            self.forum_id = forum.id
            self._forum = forum

    @property
    def submitter(self):

        if self.api and self.submitter_id:
            return self.api._get_user(self.submitter_id)

    @submitter.setter
    def submitter(self, submitter):
        if submitter:
            self.submitter_id = submitter.id
            self._submitter = submitter

    @property
    def updated(self):

        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated

    @property
    def updater(self):

        if self.api and self.updater_id:
            return self.api._get_user(self.updater_id)

    @updater.setter
    def updater(self, updater):
        if updater:
            self.updater_id = updater.id
            self._updater = updater


class MacroResult(BaseObject):
    def __init__(self, api=None, ticket=None, **kwargs):
        self.api = api
        self.ticket = ticket

        for key, value in kwargs.items():
            setattr(self, key, value)


class Activity(BaseObject):
    def __init__(self, api=None, actor=None, created_at=None, id=None, title=None, updated_at=None, url=None, user=None,
                 verb=None, **kwargs):

        self.api = api
        self.actor = actor
        self.created_at = created_at
        self.id = id
        self.title = title
        self.updated_at = updated_at
        self.url = url
        self.user = user
        self.verb = verb

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):

        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):

        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Status(BaseObject):
    def __init__(self, api=None, action=None, errors=None, id=None, status=None, success=None, title=None, **kwargs):
        self.api = api
        self.action = action
        self.errors = errors
        self.id = id
        self.status = status
        self.success = success
        self.title = title

        for key, value in kwargs.items():
            setattr(self, key, value)


class TicketSharingEvent(BaseObject):
    def __init__(self, api=None, action=None, agreement_id=None, id=None, type=None, **kwargs):

        self.api = api
        self.action = action
        self.agreement_id = agreement_id
        self.id = id
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def agreement(self):

        if self.api and self.agreement_id:
            return self.api._get_agreement(self.agreement_id)

    @agreement.setter
    def agreement(self, agreement):
        if agreement:
            self.agreement_id = agreement.id
            self._agreement = agreement


class PushEvent(BaseObject):
    def __init__(self, api=None, id=None, type=None, value=None, value_reference=None, **kwargs):
        self.api = api
        self.id = id
        self.type = type
        self.value = value
        self.value_reference = value_reference

        for key, value in kwargs.items():
            setattr(self, key, value)


class Action(BaseObject):
    def __init__(self, api=None, field=None, value=None, **kwargs):
        self.api = api
        self.field = field
        self.value = value

        for key, value in kwargs.items():
            setattr(self, key, value)


class ErrorEvent(BaseObject):
    def __init__(self, api=None, id=None, message=None, type=None, **kwargs):
        self.api = api
        self.id = id
        self.message = message
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)


class LogmeinTranscriptEvent(BaseObject):
    def __init__(self, api=None, body=None, id=None, type=None, **kwargs):
        self.api = api
        self.body = body
        self.id = id
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)


class OrganizationActivityEvent(BaseObject):
    def __init__(self, api=None, body=None, id=None, subject=None, type=None, **kwargs):

        self.api = api

        self._recipients = None

        self._via = None
        self.body = body
        self.id = id
        self.subject = subject
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def recipients(self):

        if self.api and self._recipients:
            return self.api._get_users(self._recipients)

    @recipients.setter
    def recipients(self, recipients):
        if recipients:
            self._recipients = recipients

    @property
    def via(self):

        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via


class TicketAudit(BaseObject):
    def __init__(self, api=None, audit=None, ticket=None, **kwargs):
        self.api = api
        self.audit = audit
        self.ticket = ticket

        for key, value in kwargs.items():
            setattr(self, key, value)


class Upload(BaseObject):
    def __init__(self, api=None, expires_at=None, token=None, **kwargs):

        self.api = api

        self._attachment = None

        self._attachments = None
        self.expires_at = expires_at
        self.token = token

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def attachment(self):

        if self.api and self._attachment:
            return self.api._get_attachment(self._attachment)

    @attachment.setter
    def attachment(self, attachment):
        if attachment:
            self._attachment = attachment

    @property
    def attachments(self):

        if self.api and self._attachments:
            return self.api._get_attachments(self._attachments)

    @attachments.setter
    def attachments(self, attachments):
        if attachments:
            self._attachments = attachments

    @property
    def expires(self):

        if self.expires_at:
            return dateutil.parser.parse(self.expires_at)

    @expires.setter
    def expires(self, expires):
        if expires:
            self.expires_at = expires


class User(BaseObject):
    def __init__(self, api=None, active=None, alias=None, chat_only=None, created_at=None, custom_role_id=None,
                 details=None, email=None, external_id=None, id=None, last_login_at=None, locale=None, locale_id=None,
                 moderator=None, name=None, notes=None, only_private_comments=None, organization_id=None, phone=None,
                 photo=None, restricted_agent=None, role=None, shared=None, shared_agent=None, signature=None,
                 suspended=None, tags=None, ticket_restriction=None, time_zone=None, two_factor_auth_enabled=None,
                 updated_at=None, url=None, verified=None, **kwargs):

        self.api = api

        #:| Comment: Custom fields for the user
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: :class:`hash`

        self._user_fields = None

        #:| Comment: false if the user has been deleted
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.active = active

        #:| Comment: An alias displayed to end users
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.alias = alias

        #:| Comment: Whether or not the user is a chat-only agent
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.chat_only = chat_only

        #:| Comment: The time the user was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: A custom role if the user is an agent on the Enterprise plan
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.custom_role_id = custom_role_id

        #:| Comment: Any details you want to store about the user, such as an address
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.details = details

        #:| Comment: The user's primary email address
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.email = email

        #:| Comment: A unique id you can specify for the user
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.external_id = external_id

        #:| Comment: Automatically assigned when the user is created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The last time the user signed in to Zendesk Support
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.last_login_at = last_login_at

        #:| Comment: The user's locale
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.locale = locale

        #:| Comment: The user's language identifier
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.locale_id = locale_id

        #:| Comment: Designates whether the user has forum moderation capabilities
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.moderator = moderator

        #:| Comment: The user's name
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.name = name

        #:| Comment: Any notes you want to store about the user
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.notes = notes

        #:| Comment: true if the user can only create private comments
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.only_private_comments = only_private_comments

        #:| Comment: The id of the organization the user is associated with
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.organization_id = organization_id

        #:| Comment: The user's primary phone number. See Phone Number below
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.phone = phone

        #:| Comment: The user's profile picture represented as an Attachment object
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: :class:`Attachment`
        self.photo = photo

        #:| Comment: If the agent has any restrictions; false for admins and unrestricted agents, true for other agents
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.restricted_agent = restricted_agent

        #:| Comment: The user's role. Possible values are "end-user", "agent", or "admin"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.role = role

        #:| Comment: If the user is shared from a different Zendesk Support instance. Ticket sharing accounts only
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.shared = shared

        #:| Comment: If the user is a shared agent from a different Zendesk Support instance. Ticket sharing accounts only
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.shared_agent = shared_agent

        #:| Comment: The user's signature. Only agents and admins can have signatures
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.signature = signature

        #:| Comment: If the agent is suspended. Tickets from suspended users are also suspended, and these users cannot sign in to the end user portal
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.suspended = suspended

        #:| Comment: The user's tags. Only present if your account has user tagging enabled
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: array
        self.tags = tags

        #:| Comment: Specifies which tickets the user has access to. Possible values are: "organization", "groups", "assigned", "requested", null
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.ticket_restriction = ticket_restriction

        #:| Comment: The user's time zone. See Time Zone below
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.time_zone = time_zone

        #:| Comment: If two factor authentication is enabled.
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.two_factor_auth_enabled = two_factor_auth_enabled

        #:| Comment: The time the user was last updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The user's API url
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        #:| Comment: If the user's identity has been verified or not
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.verified = verified

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def user_fields(self):
        """
        |  Comment: Custom fields for the user
        """
        if self.api and self._user_fields:
            return self.api._get_user_fields(self._user_fields)

    @user_fields.setter
    def user_fields(self, user_fields):
        if user_fields:
            self._user_fields = user_fields

    @property
    def created(self):
        """
        |  Comment: The time the user was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def custom_role(self):
        """
        |  Comment: A custom role if the user is an agent on the Enterprise plan
        """
        if self.api and self.custom_role_id:
            return self.api._get_custom_role(self.custom_role_id)

    @custom_role.setter
    def custom_role(self, custom_role):
        if custom_role:
            self.custom_role_id = custom_role.id
            self._custom_role = custom_role

    @property
    def last_login(self):
        """
        |  Comment: The last time the user signed in to Zendesk Support
        """
        if self.last_login_at:
            return dateutil.parser.parse(self.last_login_at)

    @last_login.setter
    def last_login(self, last_login):
        if last_login:
            self.last_login_at = last_login

    @property
    def organization(self):
        """
        |  Comment: The id of the organization the user is associated with
        """
        if self.api and self.organization_id:
            return self.api._get_organization(self.organization_id)

    @organization.setter
    def organization(self, organization):
        if organization:
            self.organization_id = organization.id
            self._organization = organization

    @property
    def updated(self):
        """
        |  Comment: The time the user was last updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class FacebookCommentEvent(BaseObject):
    def __init__(self, api=None, author_id=None, body=None, graph_object_id=None, html_body=None, id=None, public=None,
                 trusted=None, type=None, **kwargs):

        self.api = api

        self._attachments = None

        self._data = None
        self.author_id = author_id
        self.body = body
        self.graph_object_id = graph_object_id
        self.html_body = html_body
        self.id = id
        self.public = public
        self.trusted = trusted
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def attachments(self):

        if self.api and self._attachments:
            return self.api._get_attachments(self._attachments)

    @attachments.setter
    def attachments(self, attachments):
        if attachments:
            self._attachments = attachments

    @property
    def data(self):

        if self.api and self._data:
            return self.api._get_data(self._data)

    @data.setter
    def data(self, data):
        if data:
            self._data = data

    @property
    def author(self):

        if self.api and self.author_id:
            return self.api._get_user(self.author_id)

    @author.setter
    def author(self, author):
        if author:
            self.author_id = author.id
            self._author = author


class SatisfactionRating(BaseObject):
    def __init__(self, api=None, assignee_id=None, created_at=None, group_id=None, id=None, requester_id=None,
                 score=None, ticket_id=None, updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: The id of agent assigned to at the time of rating
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: integer
        self.assignee_id = assignee_id

        #:| Comment: The time the satisfaction rating got created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: The id of group assigned to at the time of rating
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: integer
        self.group_id = group_id

        #:| Comment: Automatically assigned upon creation
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The id of ticket requester submitting the rating
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: integer
        self.requester_id = requester_id

        #:| Comment: The rating: "offered", "unoffered", "good" or "bad"
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: string
        self.score = score

        #:| Comment: The id of ticket being rated
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: integer
        self.ticket_id = ticket_id

        #:| Comment: The time the satisfaction rating got updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this rating
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def assignee(self):
        """
        |  Comment: The id of agent assigned to at the time of rating
        """
        if self.api and self.assignee_id:
            return self.api._get_user(self.assignee_id)

    @assignee.setter
    def assignee(self, assignee):
        if assignee:
            self.assignee_id = assignee.id
            self._assignee = assignee

    @property
    def created(self):
        """
        |  Comment: The time the satisfaction rating got created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def group(self):
        """
        |  Comment: The id of group assigned to at the time of rating
        """
        if self.api and self.group_id:
            return self.api._get_group(self.group_id)

    @group.setter
    def group(self, group):
        if group:
            self.group_id = group.id
            self._group = group

    @property
    def requester(self):
        """
        |  Comment: The id of ticket requester submitting the rating
        """
        if self.api and self.requester_id:
            return self.api._get_user(self.requester_id)

    @requester.setter
    def requester(self, requester):
        if requester:
            self.requester_id = requester.id
            self._requester = requester

    @property
    def ticket(self):
        """
        |  Comment: The id of ticket being rated
        """
        if self.api and self.ticket_id:
            return self.api._get_ticket(self.ticket_id)

    @ticket.setter
    def ticket(self, ticket):
        if ticket:
            self.ticket_id = ticket.id
            self._ticket = ticket

    @property
    def updated(self):
        """
        |  Comment: The time the satisfaction rating got updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class SatisfactionRatingEvent(BaseObject):
    def __init__(self, api=None, assignee_id=None, body=None, id=None, score=None, type=None, **kwargs):

        self.api = api
        self.assignee_id = assignee_id
        self.body = body
        self.id = id
        self.score = score
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def assignee(self):

        if self.api and self.assignee_id:
            return self.api._get_user(self.assignee_id)

    @assignee.setter
    def assignee(self, assignee):
        if assignee:
            self.assignee_id = assignee.id
            self._assignee = assignee


class Comment(BaseObject):
    def __init__(self, api=None, author_id=None, body=None, created_at=None, id=None, public=None, type=None, **kwargs):

        self.api = api

        self._attachments = None

        self._metadata = None

        self._via = None
        self.author_id = author_id
        self.body = body
        self.created_at = created_at
        self.id = id
        self.public = public
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def attachments(self):

        if self.api and self._attachments:
            return self.api._get_attachments(self._attachments)

    @attachments.setter
    def attachments(self, attachments):
        if attachments:
            self._attachments = attachments

    @property
    def metadata(self):

        if self.api and self._metadata:
            return self.api._get_metadata(self._metadata)

    @metadata.setter
    def metadata(self, metadata):
        if metadata:
            self._metadata = metadata

    @property
    def via(self):

        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via

    @property
    def author(self):

        if self.api and self.author_id:
            return self.api._get_user(self.author_id)

    @author.setter
    def author(self, author):
        if author:
            self.author_id = author.id
            self._author = author

    @property
    def created(self):

        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created


class Source(BaseObject):
    def __init__(self, api=None, rel=None, **kwargs):

        self.api = api

        self._from_ = None

        self._to = None
        self.rel = rel

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def from_(self):

        if self.api and self._from_:
            return self.api._get_from_(self._from_)

    @from_.setter
    def from_(self, from_):
        if from_:
            self._from_ = from_


class ExternalEvent(BaseObject):
    def __init__(self, api=None, body=None, id=None, resource=None, type=None, **kwargs):
        self.api = api
        self.body = body
        self.id = id
        self.resource = resource
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)


class Group(BaseObject):
    def __init__(self, api=None, created_at=None, deleted=None, id=None, name=None, updated_at=None, url=None,
                 **kwargs):

        self.api = api

        #:| Comment: The time the group was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Deleted groups get marked as such
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.deleted = deleted

        #:| Comment: Automatically assigned when creating groups
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The name of the group
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.name = name

        #:| Comment: The time of the last update of the group
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this group
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: The time the group was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the group
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class SharingAgreement(BaseObject):
    def __init__(self, api=None, created_at=None, id=None, name=None, partner_name=None, remote_subdomain=None,
                 status=None, type=None, **kwargs):

        self.api = api

        #:| Comment: The time the record was created
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Automatically assigned upon creation
        #:| Type: integer
        self.id = id

        #:| Comment: Name of this sharing agreement
        #:| Type: string
        self.name = name

        #:| Comment: Can be one of the following: 'jira', null
        #:| Type: string
        self.partner_name = partner_name

        #:| Comment: Subdomain of the remote account or null if not associated with an account
        #:| Type: string
        self.remote_subdomain = remote_subdomain

        #:| Comment: Can be one of the following: 'accepted', 'declined', 'pending', 'inactive'
        #:| Type: string
        self.status = status

        #:| Comment: Can be one of the following: 'inbound', 'outbound'
        #:| Type: string
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: The time the record was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created


class CommentPrivacyChangeEvent(BaseObject):
    def __init__(self, api=None, comment_id=None, id=None, public=None, type=None, **kwargs):

        self.api = api
        self.comment_id = comment_id
        self.id = id
        self.public = public
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def comment(self):

        if self.api and self.comment_id:
            return self.api._get_comment(self.comment_id)

    @comment.setter
    def comment(self, comment):
        if comment:
            self.comment_id = comment.id
            self._comment = comment


class GroupMembership(BaseObject):
    def __init__(self, api=None, created_at=None, default=None, group_id=None, id=None, updated_at=None, url=None,
                 user_id=None, **kwargs):

        self.api = api

        #:| Comment: The time the membership was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: If true, tickets assigned directly to the agent will assume this membership's group.
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.default = default

        #:| Comment: The id of a group
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: integer
        self.group_id = group_id

        #:| Comment: Automatically assigned upon creation
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The time of the last update of the membership
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this record
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        #:| Comment: The id of an agent
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: integer
        self.user_id = user_id

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: The time the membership was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def group(self):
        """
        |  Comment: The id of a group
        """
        if self.api and self.group_id:
            return self.api._get_group(self.group_id)

    @group.setter
    def group(self, group):
        if group:
            self.group_id = group.id
            self._group = group

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the membership
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated

    @property
    def user(self):
        """
        |  Comment: The id of an agent
        """
        if self.api and self.user_id:
            return self.api._get_user(self.user_id)

    @user.setter
    def user(self, user):
        if user:
            self.user_id = user.id
            self._user = user


class OrganizationMembership(BaseObject):
    def __init__(self, api=None, created_at=None, default=None, id=None, organization_id=None, updated_at=None,
                 url=None, user_id=None, **kwargs):

        self.api = api

        #:| Comment: When this record was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Denotes whether this is the default organization membership for the user
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: boolean
        self.default = default

        #:| Comment: Automatically assigned when creating memberships
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The ID of the organization associated with this user, in this membership
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: integer
        self.organization_id = organization_id

        #:| Comment: When this record last got updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this membership
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        #:| Comment: The ID of the user for whom this memberships belongs
        #:| Mandatory: yes
        #:| Read-only: yes
        #:| Type: integer
        self.user_id = user_id

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def created(self):
        """
        |  Comment: When this record was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def organization(self):
        """
        |  Comment: The ID of the organization associated with this user, in this membership
        """
        if self.api and self.organization_id:
            return self.api._get_organization(self.organization_id)

    @organization.setter
    def organization(self, organization):
        if organization:
            self.organization_id = organization.id
            self._organization = organization

    @property
    def updated(self):
        """
        |  Comment: When this record last got updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated

    @property
    def user(self):
        """
        |  Comment: The ID of the user for whom this memberships belongs
        """
        if self.api and self.user_id:
            return self.api._get_user(self.user_id)

    @user.setter
    def user(self, user):
        if user:
            self.user_id = user.id
            self._user = user


class CreateEvent(BaseObject):
    def __init__(self, api=None, field_name=None, id=None, type=None, value=None, **kwargs):
        self.api = api
        self.field_name = field_name
        self.id = id
        self.type = type
        self.value = value

        for key, value in kwargs.items():
            setattr(self, key, value)


class TicketMetricItem(BaseObject):
    def __init__(self, api=None, business=None, calendar=None, **kwargs):
        self.api = api
        self.business = business
        self.calendar = calendar

        for key, value in kwargs.items():
            setattr(self, key, value)


class FacebookEvent(BaseObject):
    def __init__(self, api=None, body=None, communication=None, id=None, ticket_via=None, type=None, **kwargs):

        self.api = api

        self._page = None
        self.body = body
        self.communication = communication
        self.id = id
        self.ticket_via = ticket_via
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def page(self):

        if self.api and self._page:
            return self.api._get_page(self._page)

    @page.setter
    def page(self, page):
        if page:
            self._page = page


class Brand(BaseObject):
    def __init__(self, api=None, active=None, brand_url=None, created_at=None, default=None, has_help_center=None,
                 help_center_state=None, host_mapping=None, id=None, name=None, subdomain=None, updated_at=None,
                 url=None, **kwargs):

        self.api = api

        #:| Comment: Logo image for this brand
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: :class:`Attachment`

        self._logo = None

        #:| Comment: If the brand is set as active
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.active = active

        #:| Comment: The url of the brand
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.brand_url = brand_url

        #:| Comment: The time the brand was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Is the brand the default brand for this account
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.default = default

        #:| Comment: If the brand has a Help Center
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.has_help_center = has_help_center

        #:| Comment: The state of the Help Center: enabled, disabled, or restricted
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.help_center_state = help_center_state

        #:| Comment: The hostmapping to this brand, if any (only admins view this key)
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.host_mapping = host_mapping

        #:| Comment: Automatically assigned when the brand is created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The name of the brand
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.name = name

        #:| Comment: The subdomain of the brand (only admins view this key)
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.subdomain = subdomain

        #:| Comment: The time of the last update of the brand
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this brand
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def logo(self):
        """
        |  Comment: Logo image for this brand
        """
        if self.api and self._logo:
            return self.api._get_logo(self._logo)

    @logo.setter
    def logo(self, logo):
        if logo:
            self._logo = logo

    @property
    def created(self):
        """
        |  Comment: The time the brand was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the brand
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Audit(BaseObject):
    def __init__(self, api=None, author_id=None, created_at=None, id=None, ticket_id=None, **kwargs):

        self.api = api

        self._events = None

        self._metadata = None

        self._via = None
        self.author_id = author_id
        self.created_at = created_at
        self.id = id
        self.ticket_id = ticket_id

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def events(self):

        if self.api and self._events:
            return self.api._get_events(self._events)

    @events.setter
    def events(self, events):
        if events:
            self._events = events

    @property
    def metadata(self):

        if self.api and self._metadata:
            return self.api._get_metadata(self._metadata)

    @metadata.setter
    def metadata(self, metadata):
        if metadata:
            self._metadata = metadata

    @property
    def via(self):

        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via

    @property
    def author(self):

        if self.api and self.author_id:
            return self.api._get_user(self.author_id)

    @author.setter
    def author(self, author):
        if author:
            self.author_id = author.id
            self._author = author

    @property
    def created(self):

        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def ticket(self):

        if self.api and self.ticket_id:
            return self.api._get_ticket(self.ticket_id)

    @ticket.setter
    def ticket(self, ticket):
        if ticket:
            self.ticket_id = ticket.id
            self._ticket = ticket


class Thumbnail(BaseObject):
    def __init__(self, api=None, content_type=None, content_url=None, file_name=None, id=None, size=None, **kwargs):
        self.api = api
        self.content_type = content_type
        self.content_url = content_url
        self.file_name = file_name
        self.id = id
        self.size = size

        for key, value in kwargs.items():
            setattr(self, key, value)


class Macro(BaseObject):
    def __init__(self, api=None, active=None, created_at=None, description=None, id=None, position=None,
                 restriction=None, title=None, updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: An object describing what the macro will do
        #:| Type: :class:`Actions`

        self._actions = None

        #:| Comment: Useful for determining if the macro should be displayed
        #:| Type: boolean
        self.active = active

        #:| Comment: The time the macro was created
        #:| Type: date
        self.created_at = created_at

        #:| Comment: The description of the macro
        #:| Type: string
        self.description = description

        #:| Comment: Automatically assigned when created
        #:| Type: integer
        self.id = id

        #:| Comment: The position of the macro
        #:| Type: integer
        self.position = position

        #:| Comment: Who may access this macro. Will be null when everyone in the account can access it
        #:| Type: object
        self.restriction = restriction

        #:| Comment: The title of the macro
        #:| Type: string
        self.title = title

        #:| Comment: The time of the last update of the macro
        #:| Type: date
        self.updated_at = updated_at
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def actions(self):
        """
        |  Comment: An object describing what the macro will do
        """
        if self.api and self._actions:
            return self.api._get_actions(self._actions)

    @actions.setter
    def actions(self, actions):
        if actions:
            self._actions = actions

    @property
    def created(self):
        """
        |  Comment: The time the macro was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the macro
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Attachment(BaseObject):
    def __init__(self, api=None, content_type=None, content_url=None, file_name=None, id=None, size=None, **kwargs):

        self.api = api

        #:| Comment: An array of Photo objects. Note that thumbnails do not have thumbnails.
        #:| Read-only: yes
        #:| Type: array

        self._thumbnails = None

        #:| Comment: The content type of the image. Example value: image/png
        #:| Read-only: yes
        #:| Type: string
        self.content_type = content_type

        #:| Comment: A full URL where the attachment image file can be downloaded
        #:| Read-only: yes
        #:| Type: string
        self.content_url = content_url

        #:| Comment: The name of the image file
        #:| Read-only: yes
        #:| Type: string
        self.file_name = file_name

        #:| Comment: Automatically assigned when created
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The size of the image file in bytes
        #:| Read-only: yes
        #:| Type: integer
        self.size = size

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def thumbnails(self):
        """
        |  Comment: An array of Photo objects. Note that thumbnails do not have thumbnails.
        """
        if self.api and self._thumbnails:
            return self.api._get_thumbnails(self._thumbnails)

    @thumbnails.setter
    def thumbnails(self, thumbnails):
        if thumbnails:
            self._thumbnails = thumbnails


class SuspendedTicket(BaseObject):
    def __init__(self, api=None, brand_id=None, cause=None, content=None, created_at=None, id=None, recipient=None,
                 subject=None, ticket_id=None, updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: The author id (if available), name and email
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._author = None

        #:| Comment: This object explains how the ticket was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: :class:`Via`

        self._via = None

        #:| Comment: The id of the brand this ticket is associated with - only applicable for enterprise accounts
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.brand_id = brand_id

        #:| Comment: Why the ticket was suspended
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.cause = cause

        #:| Comment: The content that was flagged
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.content = content

        #:| Comment: When this record was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Automatically assigned
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The original recipient e-mail address of the ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.recipient = recipient

        #:| Comment: The value of the subject field for this ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.subject = subject

        #:| Comment: The ticket ID this suspended email is associated with, if available
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.ticket_id = ticket_id

        #:| Comment: When this record last got updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def via(self):
        """
        |  Comment: This object explains how the ticket was created
        """
        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via

    @property
    def brand(self):
        """
        |  Comment: The id of the brand this ticket is associated with - only applicable for enterprise accounts
        """
        if self.api and self.brand_id:
            return self.api._get_brand(self.brand_id)

    @brand.setter
    def brand(self, brand):
        if brand:
            self.brand_id = brand.id
            self._brand = brand

    @property
    def created(self):
        """
        |  Comment: When this record was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def ticket(self):
        """
        |  Comment: The ticket ID this suspended email is associated with, if available
        """
        if self.api and self.ticket_id:
            return self.api._get_ticket(self.ticket_id)

    @ticket.setter
    def ticket(self, ticket):
        if ticket:
            self.ticket_id = ticket.id
            self._ticket = ticket

    @property
    def updated(self):
        """
        |  Comment: When this record last got updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class ChangeEvent(BaseObject):
    def __init__(self, api=None, field_name=None, id=None, previous_value=None, type=None, value=None, **kwargs):
        self.api = api
        self.field_name = field_name
        self.id = id
        self.previous_value = previous_value
        self.type = type
        self.value = value

        for key, value in kwargs.items():
            setattr(self, key, value)


class JobStatus(BaseObject):
    def __init__(self, api=None, id=None, message=None, progress=None, results=None, status=None, total=None, url=None,
                 **kwargs):
        self.api = api
        self.id = id
        self.message = message
        self.progress = progress
        self.results = results
        self.status = status
        self.total = total
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)


class CcEvent(BaseObject):
    def __init__(self, api=None, id=None, type=None, **kwargs):

        self.api = api

        self._recipients = None

        self._via = None
        self.id = id
        self.type = type

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def recipients(self):

        if self.api and self._recipients:
            return self.api._get_users(self._recipients)

    @recipients.setter
    def recipients(self, recipients):
        if recipients:
            self._recipients = recipients

    @property
    def via(self):

        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via


class CustomField(BaseObject):
    def __init__(self, api=None, id=None, value=None, **kwargs):
        self.api = api
        self.id = id
        self.value = value

        for key, value in kwargs.items():
            setattr(self, key, value)


class Organization(BaseObject):
    def __init__(self, api=None, created_at=None, details=None, external_id=None, group_id=None, id=None, name=None,
                 notes=None, shared_comments=None, shared_tickets=None, tags=None, updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: An array of domain names associated with this organization
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: array

        self._domain_names = None

        #:| Comment: Custom fields for this organization
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: :class:`hash`

        self._organization_fields = None

        #:| Comment: The time the organization was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: Any details obout the organization, such as the address
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.details = details

        #:| Comment: A unique external id to associate organizations to an external record
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.external_id = external_id

        #:| Comment: New tickets from users in this organization are automatically put in this group
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.group_id = group_id

        #:| Comment: Automatically assigned when the organization is created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: A unique name for the organization
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: string
        self.name = name

        #:| Comment: Any notes you have about the organization
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.notes = notes

        #:| Comment: End users in this organization are able to see each other's comments on tickets
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.shared_comments = shared_comments

        #:| Comment: End users in this organization are able to see each other's tickets
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: boolean
        self.shared_tickets = shared_tickets

        #:| Comment: The tags of the organization
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: array
        self.tags = tags

        #:| Comment: The time of the last update of the organization
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this organization
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def organization_fields(self):
        """
        |  Comment: Custom fields for this organization
        """
        if self.api and self._organization_fields:
            return self.api._get_organization_fields(self._organization_fields)

    @organization_fields.setter
    def organization_fields(self, organization_fields):
        if organization_fields:
            self._organization_fields = organization_fields

    @property
    def created(self):
        """
        |  Comment: The time the organization was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def group(self):
        """
        |  Comment: New tickets from users in this organization are automatically put in this group
        """
        if self.api and self.group_id:
            return self.api._get_group(self.group_id)

    @group.setter
    def group(self, group):
        if group:
            self.group_id = group.id
            self._group = group

    @property
    def updated(self):
        """
        |  Comment: The time of the last update of the organization
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Ticket(BaseObject):
    def __init__(self, api=None, assignee_id=None, brand_id=None, collaborator_ids=None, created_at=None,
                 description=None, due_at=None, external_id=None, forum_topic_id=None, group_id=None,
                 has_incidents=None, id=None, organization_id=None, priority=None, problem_id=None, raw_subject=None,
                 recipient=None, requester_id=None, sharing_agreement_ids=None, status=None, subject=None,
                 submitter_id=None, tags=None, type=None, updated_at=None, url=None, **kwargs):

        self.api = api

        #:| Comment: The custom fields of the ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: array

        self._custom_fields = None

        self._fields = None

        #:| Comment: The satisfaction rating of the ticket, if it exists, or the state of satisfaction, 'offered' or 'unoffered'
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: object

        self._satisfaction_rating = None

        #:| Comment: This object explains how the ticket was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: :class:`Via`

        self._via = None

        #:| Comment: What agent is currently assigned to the ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.assignee_id = assignee_id

        #:| Comment: The id of the brand this ticket is associated with - only applicable for enterprise accounts
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.brand_id = brand_id

        #:| Comment: Who are currently CC'ed on the ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: array
        self.collaborator_ids = collaborator_ids

        #:| Comment: When this record was created
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.created_at = created_at

        #:| Comment: The first comment on the ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.description = description

        #:| Comment: If this is a ticket of type "task" it has a due date.  Due date format uses ISO 8601 format.
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: date
        self.due_at = due_at

        #:| Comment: An id you can use to link Zendesk Support tickets to local records
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.external_id = external_id

        #:| Comment: The topic this ticket originated from, if any
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.forum_topic_id = forum_topic_id

        #:| Comment: The group this ticket is assigned to
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.group_id = group_id

        #:| Comment: Is true of this ticket has been marked as a problem, false otherwise
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: boolean
        self.has_incidents = has_incidents

        #:| Comment: Automatically assigned when creating tickets
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: integer
        self.id = id

        #:| Comment: The organization of the requester. *Writable only when creating, not when updating
        #:| Mandatory: no
        #:| Read-only: no*
        #:| Type: integer
        self.organization_id = organization_id

        #:| Comment: Priority, defines the urgency with which the ticket should be addressed: "urgent", "high", "normal", "low"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.priority = priority

        #:| Comment: The problem this incident is linked to, if any
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.problem_id = problem_id

        #:| Comment: The dynamic content placeholder, if present, or the "subject" value, if not. See Dynamic Content
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.raw_subject = raw_subject

        #:| Comment: The original recipient e-mail address of the ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.recipient = recipient

        #:| Comment: The user who requested this ticket
        #:| Mandatory: yes
        #:| Read-only: no
        #:| Type: integer
        self.requester_id = requester_id

        #:| Comment: The ids of the sharing agreements used for this ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: array
        self.sharing_agreement_ids = sharing_agreement_ids

        #:| Comment: The state of the ticket, "new", "open", "pending", "hold", "solved", "closed"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.status = status

        #:| Comment: The value of the subject field for this ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.subject = subject

        #:| Comment: The user who submitted the ticket; The submitter always becomes the author of the first comment on the ticket.
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: integer
        self.submitter_id = submitter_id

        #:| Comment: The array of tags applied to this ticket
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: array
        self.tags = tags

        #:| Comment: The type of this ticket, i.e. "problem", "incident", "question" or "task"
        #:| Mandatory: no
        #:| Read-only: no
        #:| Type: string
        self.type = type

        #:| Comment: When this record last got updated
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: date
        self.updated_at = updated_at

        #:| Comment: The API url of this ticket
        #:| Mandatory: no
        #:| Read-only: yes
        #:| Type: string
        self.url = url

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def custom_fields(self):
        """
        |  Comment: The custom fields of the ticket
        """
        if self.api and self._custom_fields:
            return self.api._get_custom_fields(self._custom_fields)

    @custom_fields.setter
    def custom_fields(self, custom_fields):
        if custom_fields:
            self._custom_fields = custom_fields

    @property
    def fields(self):

        if self.api and self._fields:
            return self.api._get_fields(self._fields)

    @fields.setter
    def fields(self, fields):
        if fields:
            self._fields = fields

    @property
    def satisfaction_rating(self):
        """
        |  Comment: The satisfaction rating of the ticket, if it exists, or the state of satisfaction, 'offered' or 'unoffered'
        """
        if self.api and self._satisfaction_rating:
            return self.api._get_satisfaction_rating(self._satisfaction_rating)

    @satisfaction_rating.setter
    def satisfaction_rating(self, satisfaction_rating):
        if satisfaction_rating:
            self._satisfaction_rating = satisfaction_rating

    @property
    def via(self):
        """
        |  Comment: This object explains how the ticket was created
        """
        if self.api and self._via:
            return self.api._get_via(self._via)

    @via.setter
    def via(self, via):
        if via:
            self._via = via

    @property
    def assignee(self):
        """
        |  Comment: What agent is currently assigned to the ticket
        """
        if self.api and self.assignee_id:
            return self.api._get_user(self.assignee_id)

    @assignee.setter
    def assignee(self, assignee):
        if assignee:
            self.assignee_id = assignee.id
            self._assignee = assignee

    @property
    def brand(self):
        """
        |  Comment: The id of the brand this ticket is associated with - only applicable for enterprise accounts
        """
        if self.api and self.brand_id:
            return self.api._get_brand(self.brand_id)

    @brand.setter
    def brand(self, brand):
        if brand:
            self.brand_id = brand.id
            self._brand = brand

    @property
    def collaborators(self):
        """
        |  Comment: Who are currently CC'ed on the ticket
        """
        if self.api and self.collaborator_ids:
            return self.api._get_users(self.collaborator_ids)

    @collaborators.setter
    def collaborators(self, collaborators):
        if collaborators:
            self.collaborator_ids = [o.id for o in collaborators]
            self._collaborators = collaborators

    @property
    def created(self):
        """
        |  Comment: When this record was created
        """
        if self.created_at:
            return dateutil.parser.parse(self.created_at)

    @created.setter
    def created(self, created):
        if created:
            self.created_at = created

    @property
    def due(self):
        """
        |  Comment: If this is a ticket of type "task" it has a due date.  Due date format uses ISO 8601 format.
        """
        if self.due_at:
            return dateutil.parser.parse(self.due_at)

    @due.setter
    def due(self, due):
        if due:
            self.due_at = due

    @property
    def forum_topic(self):
        """
        |  Comment: The topic this ticket originated from, if any
        """
        if self.api and self.forum_topic_id:
            return self.api._get_topic(self.forum_topic_id)

    @forum_topic.setter
    def forum_topic(self, forum_topic):
        if forum_topic:
            self.forum_topic_id = forum_topic.id
            self._forum_topic = forum_topic

    @property
    def group(self):
        """
        |  Comment: The group this ticket is assigned to
        """
        if self.api and self.group_id:
            return self.api._get_group(self.group_id)

    @group.setter
    def group(self, group):
        if group:
            self.group_id = group.id
            self._group = group

    @property
    def organization(self):
        """
        |  Comment: The organization of the requester. *Writable only when creating, not when updating
        """
        if self.api and self.organization_id:
            return self.api._get_organization(self.organization_id)

    @organization.setter
    def organization(self, organization):
        if organization:
            self.organization_id = organization.id
            self._organization = organization

    @property
    def problem(self):
        """
        |  Comment: The problem this incident is linked to, if any
        """
        if self.api and self.problem_id:
            return self.api._get_problem(self.problem_id)

    @problem.setter
    def problem(self, problem):
        if problem:
            self.problem_id = problem.id
            self._problem = problem

    @property
    def requester(self):
        """
        |  Comment: The user who requested this ticket
        """
        if self.api and self.requester_id:
            return self.api._get_user(self.requester_id)

    @requester.setter
    def requester(self, requester):
        if requester:
            self.requester_id = requester.id
            self._requester = requester

    @property
    def sharing_agreements(self):
        """
        |  Comment: The ids of the sharing agreements used for this ticket
        """
        if self.api and self.sharing_agreement_ids:
            return self.api._get_sharing_agreements(self.sharing_agreement_ids)

    @sharing_agreements.setter
    def sharing_agreements(self, sharing_agreements):
        if sharing_agreements:
            self.sharing_agreement_ids = [o.id for o in sharing_agreements]
            self._sharing_agreements = sharing_agreements

    @property
    def submitter(self):
        """
        |  Comment: The user who submitted the ticket; The submitter always becomes the author of the first comment on the ticket.
        """
        if self.api and self.submitter_id:
            return self.api._get_user(self.submitter_id)

    @submitter.setter
    def submitter(self, submitter):
        if submitter:
            self.submitter_id = submitter.id
            self._submitter = submitter

    @property
    def updated(self):
        """
        |  Comment: When this record last got updated
        """
        if self.updated_at:
            return dateutil.parser.parse(self.updated_at)

    @updated.setter
    def updated(self, updated):
        if updated:
            self.updated_at = updated


class Via(BaseObject):
    def __init__(self, api=None, **kwargs):

        self.api = api

        self._source = None

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def source(self):

        if self.api and self._source:
            return self.api._get_source(self._source)

    @source.setter
    def source(self, source):
        if source:
            self._source = source



import json
import logging
import os
from time import sleep

from zenpy.lib.api_objects import User, Ticket, Macro
from zenpy.lib.endpoint import Endpoint
from zenpy.lib.exception import APIException, RecordNotFoundException
from zenpy.lib.exception import ZenpyException
from zenpy.lib.generator import ResultGenerator
from zenpy.lib.manager import ApiObjectEncoder, ObjectManager
from zenpy.lib.util import to_snake_case

__author__ = 'facetoe'

log = logging.getLogger(__name__)


def serialize(api_object):
    return json.loads(json.dumps(api_object, cls=ApiObjectEncoder))


class Api(object):
    """
    Base class for API.
    """
    subdomain = None

    def __init__(self, subdomain, session, endpoint, object_type, timeout):
        self.subdomain = subdomain
        self.session = session
        self.timeout = timeout
        self.endpoint = endpoint
        self.object_type = object_type
        self.protocol = 'https'
        self.version = 'v2'
        self.object_manager = ObjectManager(self)

    def post(self, url, payload, data=None):
        log.debug("POST: %s - %s" % (url, str(payload)))
        headers = None
        if data:
            headers = {'Content-Type': 'application/octet-stream'}
        response = self.session.post(url, json=serialize(payload), data=data, headers=headers, timeout=self.timeout)
        self._check_and_cache_response(response)
        return self._build_response(response.json())

    def _put(self, url, payload):
        return self._call_api(self.session.put, url, json=serialize(payload), timeout=self.timeout)

    def _delete(self, url, payload=None):
        return self._call_api(self.session.delete, url, json=payload, timeout=self.timeout)

    def _get(self, url):
        return self._call_api(self.session.get, url, timeout=self.timeout)

    def _call_api(self, http_method, url, **kwargs):
        log.debug("{}: {} - {}".format(http_method.__name__.upper(), url, kwargs))
        response = http_method(url, **kwargs)

        # If we are being rate-limited, wait the required period before trying again.
        while 'retry-after' in response.headers and int(response.headers['retry-after']) > 0:
            retry_after_seconds = int(response.headers['retry-after'])
            log.warn(
                "Waiting for requested retry-after period: %s seconds" % retry_after_seconds)
            while retry_after_seconds > 0:
                retry_after_seconds -= 1
                log.debug("    -> sleeping: %s more seconds" % retry_after_seconds)
                sleep(1)
            response = http_method(url, **kwargs)
        return self._check_and_cache_response(response)

    def _get_items(self, endpoint, object_type, *args, **kwargs):
        sideload = 'sideload' not in kwargs or ('sideload' in kwargs and kwargs['sideload'])

        # If an ID is present a single object has been requested
        if 'id' in kwargs:
            return self._get_item(kwargs['id'], endpoint, object_type, sideload)

        if 'ids' in kwargs:
            cached_objects = []
            # Check to see if we have all objects in the cache.
            # If we are missing even one we need to request them all again.
            for _id in kwargs['ids']:
                obj = self.object_manager.query_cache(object_type, _id)
                if obj:
                    cached_objects.append(obj)
                else:
                    return self._get_paginated(endpoint, object_type, *args, **kwargs)
            return cached_objects

        return self._get_paginated(endpoint, object_type, *args, **kwargs)

    def _get_item(self, _id, endpoint, object_type, sideload=True, skip_cache=False):
        if not skip_cache:
            # Check if we already have this item in the cache
            item = self.object_manager.query_cache(object_type, _id)
            if item:
                return item

        _json = self._query(endpoint=endpoint(id=_id, sideload=sideload))

        # If the result is paginated return a generator
        if 'next_page' in _json:
            return ResultGenerator(self, object_type, _json)
        # Annoyingly, tags is always plural.
        if 'tags' in _json:
            return self.object_manager.object_from_json(object_type, _json[object_type + 's'])
        else:
            return self.object_manager.object_from_json(object_type, _json[object_type])

    def _get_paginated(self, endpoint, object_type, *args, **kwargs):
        _json = self._query(endpoint=endpoint(*args, **kwargs))
        return ResultGenerator(self, object_type, _json)

    def _query(self, endpoint):
        response = self._get(self._get_url(endpoint=endpoint))
        return response.json()

    def _build_response(self, response_json):
        # When updating and deleting API objects various responses can be returned
        # We can figure out what we have by the keys in the returned JSON
        if 'ticket' and 'audit' in response_json:
            return self.object_manager.object_from_json('ticket_audit', response_json)
        elif 'tags' in response_json:
            return response_json['tags']

        known_objects = ('ticket',
                         'user',
                         'job_status',
                         'group',
                         'satisfaction_rating',
                         'request',
                         'organization',
                         'organization_membership',
                         'upload',
                         'result')

        for object_type in known_objects:
            if object_type in response_json:
                return self.object_manager.object_from_json(object_type, response_json[object_type])

        raise ZenpyException("Unknown Response: " + str(response_json))

    def _check_and_cache_response(self, response):
        if response.status_code > 299 or response.status_code < 200:
            log.debug("Received response code [%s] - headers: %s" % (response.status_code, str(response.headers)))
            # If it's just a RecordNotFound error raise the right exception,
            # otherwise try and get a nice error message.
            if 'application/json' in response.headers['content-type']:
                try:
                    _json = response.json()
                    if 'error' in _json and _json['error'] == 'RecordNotFound':
                        raise RecordNotFoundException(json.dumps(_json))
                    else:
                        raise APIException(json.dumps(_json))
                except ValueError:
                    pass

            # No can do, just raise the correct Exception.
            response.raise_for_status()
        else:
            try:
                self.object_manager.update_caches(response.json())
            except ValueError:
                pass
            return response

    def _object_from_json(self, object_type, object_json):
        return self.object_manager.object_from_json(object_type, object_json)

    def _query_cache(self, object_type, _id):
        return self.object_manager.query_cache(object_type, _id)

    def _get_url(self, endpoint=''):
        return "%(protocol)s://%(subdomain)s.zendesk.com/api/%(version)s/" % self.__dict__ + endpoint

    def __call__(self, *args, **kwargs):
        """
        Retrieve API objects. If called with no arguments returns a ResultGenerator of
        all retrievable items. Alternatively, can be called with an id to only return that item.
        """
        return self._get_items(self.endpoint, self.object_type, *args, **kwargs)

    def _get_user(self, _id):
        return self._get_item(_id, endpoint=Endpoint.users, object_type='user', sideload=True)

    def _get_users(self, _ids):
        return self._get_items(endpoint=Endpoint.users, object_type='user', ids=_ids)

    def _get_comment(self, _id):
        return self._get_item(_id, endpoint=Endpoint.tickets.comments, object_type='comment', sideload=True)

    def _get_organization(self, _id):
        return self._get_item(_id, endpoint=Endpoint.organizations, object_type='organization', sideload=True)

    def _get_group(self, _id):
        return self._get_item(_id, endpoint=Endpoint.groups, object_type='group', sideload=True)

    def _get_brand(self, _id):
        return self._get_item(_id, endpoint=Endpoint.brands, object_type='brand', sideload=True)

    def _get_ticket(self, _id, skip_cache=False):
        return self._get_item(_id, endpoint=Endpoint.tickets, object_type='ticket', sideload=False,
                              skip_cache=skip_cache)

    def _get_actions(self, actions):
        for action in actions:
            yield self._object_from_json('action', action)

    def _get_events(self, events):
        for event in events:
            yield self._object_from_json(event['type'].lower(), event)

    def _get_via(self, via):
        return self._object_from_json('via', via)

    def _get_source(self, source):
        return self._object_from_json('source', source)

    def _get_attachments(self, attachments):
        for attachment in attachments:
            yield self._object_from_json('attachment', attachment)

    def _get_thumbnails(self, thumbnails):
        for thumbnail in thumbnails:
            yield self._object_from_json('thumbnail', thumbnail)

    def _get_satisfaction_rating(self, satisfaction_rating):
        return self._object_from_json('satisfaction_rating', satisfaction_rating)

    def _get_sharing_agreements(self, sharing_agreement_ids):
        sharing_agreements = []
        for _id in sharing_agreement_ids:
            sharing_agreement = self._get_item(_id, Endpoint.sharing_agreements, 'sharing_agreement')
            if sharing_agreement:
                sharing_agreements.append(sharing_agreement)
        return sharing_agreements

    def _get_ticket_metric_item(self, metric_item):
        return self._object_from_json('ticket_metric_item', metric_item)

    def _get_metadata(self, metadata):
        return self._object_from_json('metadata', metadata)

    def _get_system(self, system):
        return self._object_from_json('system', system)

    def _get_problem(self, problem_id):
        return self._get_item(problem_id, Endpoint.tickets, 'ticket')

    # This will be deprecated soon - https://developer.zendesk.com/rest_api/docs/web-portal/forums
    def _get_forum(self, forum_id):
        return forum_id

    def _get_user_fields(self, user_fields):
        return self._get_fields(user_fields, endpoint=Endpoint.user_fields, object_type='user_field')

    def _get_organization_fields(self, organization_fields):
        return self._get_fields(organization_fields, endpoint=Endpoint().organizations.organization_fields,
                                object_type='organization_field')

    # TODO implement this with Enterprise
    def _get_custom_fields(self, custom_fields):
        return custom_fields

    # This is ticket fields, hopefully it doesn't conflict with another field type
    def _get_fields(self, fields, object_type='ticket_field', endpoint=Endpoint().ticket_fields):
        if any([self._query_cache(object_type, field) is None for field in [f['id'] for f in fields]]):
            # Populate field cache
            self._get(self._get_url(endpoint=endpoint()))
        for field in fields:
            yield self._query_cache(object_type, field)

    def _get_upload(self, upload):
        return self._object_from_json('upload', upload)

    def _get_attachment(self, attachment):
        return self._object_from_json('attachment', attachment)


class ModifiableApi(Api):
    """
    ModifiableApi contains helper methods for modifying an API
    """

    def _get_type_and_payload(self, items):
        self._check_type(items)
        if isinstance(items, list):
            first_obj = next((x for x in items))
            # Object name needs to be plural when targeting many
            object_type = "%ss" % to_snake_case(first_obj.__class__.__name__)
            payload = {object_type: [json.loads(json.dumps(i, cls=ApiObjectEncoder)) for i in items]}
        else:
            object_type = to_snake_case(items.__class__.__name__)
            payload = {object_type: json.loads(json.dumps(items, cls=ApiObjectEncoder))}
        return object_type, payload

    def _check_type(self, items):
        # We don't want people passing, for example, a Group object to a Ticket endpoint.
        expected_class = self.object_manager.class_manager.class_for_type(self.object_type)
        if isinstance(items, list):
            if any((o.__class__ is not expected_class for o in items)):
                raise ZenpyException("Invalid type - expected %(expected_class)s" % locals())
        else:
            if items.__class__ is not expected_class:
                raise ZenpyException("Invalid type - expected %(expected_class)s" % locals())

    def _do(self, action, endpoint_kwargs, payload=None, endpoint=None):
        if not endpoint:
            endpoint = self.endpoint
        url = self._get_url(endpoint=endpoint(**endpoint_kwargs))
        return action(url, payload=payload)


class CRUDApi(ModifiableApi):
    """
    CRUDApi supports create/update/delete operations
    """

    # TODO - Fix the post method to be consistent with get and put.
    def create(self, api_objects):
        """
        Create (POST) one or more API objects. Before being submitted to Zendesk the object or objects
        will be serialized to JSON.

        :param api_objects: object or objects to create
        """
        object_type, payload = self._get_type_and_payload(api_objects)
        if object_type.endswith('s'):
            return self._do(self.post, dict(create_many=True, sideload=False), payload=payload)
        else:
            return self._do(self.post, dict(sideload=False), payload=payload)

    def update(self, items):
        """
        Update (PUT) one or more API objects. Before being submitted to Zendesk the object or objects
        will be serialized to JSON.

        :param items: object or objects to update
        """
        object_type, payload = self._get_type_and_payload(items)
        if object_type.endswith('s'):
            response = self._do(self._put, dict(update_many=True, sideload=False), payload=payload)
        else:
            response = self._do(self._put, dict(id=items.id, sideload=False), payload=payload)
        return self._build_response(response_json=response.json())

    def delete(self, items):
        """
        Delete (DELETE) one or more API objects. After successfully deleting the objects from the API
        they will also be removed from the relevant Zenpy caches.

        :param items: object or objects to delete
        """
        object_type, payload = self._get_type_and_payload(items)
        if object_type.endswith('s'):
            response = self._do(self._delete, dict(destroy_ids=[i.id for i in items], sideload=False))
        else:
            response = self._do(self._delete, dict(id=items.id, sideload=False))
        self.object_manager.delete_from_cache(items)
        return response


class SuspendedTicketApi(ModifiableApi):
    """
    The SuspendedTicketApi adds some SuspendedTicket specific functionality
    """

    def recover(self, tickets):
        """
        Recover (PUT) one or more SuspendedTickets.

        :param tickets: one or more SuspendedTickets to recover
        """
        object_type, payload = self._get_type_and_payload(tickets)
        if object_type.endswith('s'):
            return self._do(self._put, dict(
                recover_ids=[i.id for i in tickets], sideload=False),
                            endpoint=self.endpoint, payload=payload)
        else:
            return self._do(self._put, dict(id=tickets.id, sideload=False),
                            endpoint=self.endpoint.recover,
                            payload=payload)

    def delete(self, tickets):
        """
        Delete (DELETE) one or more SuspendedTickets.

        :param tickets: one or more SuspendedTickets to delete
        """
        object_type, payload = self._get_type_and_payload(tickets)
        if object_type.endswith('s'):
            response = self._do(self._delete, dict(destroy_ids=[i.id for i in tickets], sideload=False))
        else:
            response = self._do(self._delete, dict(id=tickets.id, sideload=False))
        return response


# noinspection PyShadowingBuiltins
class TaggableApi(Api):
    """
    TaggableApi supports getting, setting, adding and deleting tags.
    """

    def add_tags(self, id, tags):
        """
        Add (PUT) one or more tags.

        :param id: the id of the object to tag
        :param tags: array of tags to apply to object
        """
        return self._put(self._get_url(
            endpoint=self.endpoint.tags(
                id=id,
                sideload=False)),
            payload={'tags': tags})

    def set_tags(self, id, tags):
        """
        Set (POST) one or more tags.

        :param id: the id of the object to tag
        :param tags: array of tags to apply to object
        """
        return self.post(self._get_url(
            endpoint=self.endpoint.tags(
                id=id,
                sideload=False)),
            payload={'tags': tags})

    def delete_tags(self, id, tags):
        """
        Delete (DELETE) one or more tags.

        :param id: the id of the object to delete tag from
        :param tags: array of tags to delete from object
        """
        return self._delete(self._get_url(
            endpoint=self.endpoint.tags(
                id=id,
                sideload=False, )),
            payload={'tags': tags})

    def tags(self, id):
        """
        Lists the most popular recent tags in decreasing popularity
        """
        return self._get_items(self.endpoint.tags, 'tag', id=id)


# noinspection PyShadowingBuiltins
class RateableApi(Api):
    """
    Supports rating with a SatisfactionRating
    """

    def rate(self, id, rating):
        """
        Add (POST) a satisfaction rating.

        :param id: id of object to rate
        :param rating: SatisfactionRating
        """
        return self.post(self._get_url(self.endpoint.satisfaction_ratings(
            id=id,
            sideload=False
        )),
            payload={'satisfaction_rating': vars(rating)})


class IncrementalApi(Api):
    """
    IncrementalApi supports the incremental endpoint.
    """

    def incremental(self, start_time):
        """
        Retrieve bulk data from the incremental API.
        :param start_time: The time of the oldest object you are interested in.
        """
        return self._get_items(self.endpoint.incremental, self.object_type, start_time=start_time)


class UserApi(TaggableApi, IncrementalApi, CRUDApi):
    """
    The UserApi adds some User specific functionality
    """

    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, object_type='user', timeout=timeout)

    def groups(self, user_id):
        """
        Retrieve the groups for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.groups, 'group', id=user_id)

    def organizations(self, user_id):
        """
        Retrieve the organizations for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.organizations, 'organization', id=user_id)

    def requested(self, user_id):
        """
        Retrieve the requested tickets for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.requested, 'ticket', id=user_id)

    def cced(self, user_id):
        """
        Retrieve the tickets this user is cc'd into.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.cced, 'ticket', id=user_id)

    def assigned(self, user_id):
        """
        Retrieve the assigned tickets for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.assigned, 'ticket', id=user_id)

    def group_memberships(self, user_id):
        """
        Retrieve the group memberships for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.group_memberships, 'group_membership', id=user_id)

    def requests(self, **kwargs):
        return self._get_items(self.endpoint.requests, 'request', **kwargs)

    def related(self, user_id):
        """
        Returns the UserRelated information for the requested User

        :param user_id: User id
        :return: UserRelated
        """
        return self._get_items(self.endpoint.related, 'user_related', id=user_id)

    def me(self):
        """
        Return the logged in user
        """
        return self._get_item(None, self.endpoint.me, 'user')

    def merge(self, source_user, dest_user):
        """
        Merge the user provided in source_user into dest_user

        :param source_user: User object or id of user to be merged
        :param dest_user: User object or id to merge into
        :return: The merged User
        """
        if isinstance(source_user, User):
            source_user = source_user.id
        if isinstance(dest_user, User):
            dest_user = dict(id=dest_user.id)
        else:
            dest_user = dict(id=dest_user)

        response = self._do(self._put,
                            endpoint_kwargs=dict(id=source_user),
                            payload=dict(user=dest_user),
                            endpoint=self.endpoint.merge)

        return self._build_response(response_json=response.json())

    def user_fields(self, user_id):
        """
        Retrieve the user fields for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.user_fields, 'user_field', id=user_id)

    def organization_memberships(self, user_id):
        """
        Retrieve the organization memberships for this user.

        :param user_id: user id
        """
        return self._get_items(self.endpoint.organization_memberships, 'organization_membership', id=user_id)

    def create_or_update(self, users):
        """
        Creates a user (POST) if the user does not already exist, or updates an existing user identified
        by e-mail address or external ID.

        :param users: User object or list of User objects
        :return: the created/updated User or a  JobStatus object if a list was passed
        """

        object_type, payload = self._get_type_and_payload(users)
        if object_type.endswith('s'):
            return self._do(self.post,
                            dict(create_or_update_many=True, sideload=False),
                            payload=payload,
                            endpoint=self.endpoint.create_or_update_many)
        else:
            return self._do(self.post,
                            dict(sideload=False),
                            payload=payload,
                            endpoint=self.endpoint.create_or_update)


class AttachmentApi(Api):
    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, object_type='attachment', timeout=timeout)

    def __call__(self, *args, **kwargs):
        if 'id' not in kwargs:
            raise ZenpyException("Attachment endpoint requires an id")
        return Api.__call__(self, *args, **kwargs)

    def upload(self, fp, token=None, target_name=None):
        """
        Upload a file to Zendesk.

        :param fp: file object, StringIO instance, content, or file path to be
                   uploaded
        :param token: upload token for uploading multiple files
        :param target_name: name of the file inside Zendesk
        :return: :class:`Upload` object containing a token and other information
                    (see https://developer.zendesk.com/rest_api/docs/core/attachments#uploading-files)
        """

        if hasattr(fp, 'read'):
            # File-like objects such as:
            #   PY3: io.StringIO, io.TextIOBase, io.BufferedIOBase
            #   PY2: file, io.StringIO, StringIO.StringIO, cStringIO.StringIO

            if not hasattr(fp, 'name') and not target_name:
                raise ZenpyException("upload requires a target file name")
            else:
                target_name = target_name or fp.name

        elif isinstance(fp, str):
            if os.path.isfile(fp):
                fp = open(fp, 'rb')
                target_name = target_name or fp.name
            elif not target_name:
                # Valid string, which is not a path, and without a target name
                raise ZenpyException("upload requires a target file name")

        elif not target_name:
            # Other serializable types accepted by requests (like dict)
            raise ZenpyException("upload requires a target file name")

        return self.post(self._get_url(self.endpoint.upload(filename=target_name, token=token)),
                         data=fp,
                         payload={})


class EndUserApi(CRUDApi):
    """
    EndUsers can only update.
    """

    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='user')

    def delete(self, items):
        raise ZenpyException("EndUsers cannot delete!")

    def create(self, api_objects):
        raise ZenpyException("EndUsers cannot create!")


class OrganizationApi(TaggableApi, IncrementalApi, CRUDApi):
    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='organization')

    def organization_fields(self, org_id):
        """
        Retrieve the organization fields for this organization.

        :param org_id: organization id
        """
        return self._get_items(self.endpoint.organization_fields, 'organization_field', id=org_id)

    def organization_memberships(self, org_id):
        """
        Retrieve the organization fields for this organization.

        :param org_id: organization id
        """
        return self._get_items(self.endpoint.organization_memberships, 'organization_membership', id=org_id)

    def external(self, external_id):
        """
        Locate an Organization by it's external_id attribute.

        :param external_id: external id of organization
        """
        return self._get_items(self.endpoint.external, 'organization', id=external_id)

    def requests(self, **kwargs):
        return self._get_items(self.endpoint.requests, 'request', **kwargs)


class OrganizationMembershipApi(CRUDApi):
    """
    The OrganizationMembershipApi allows the creation and deletion of Organization Memberships
    """

    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='organization_membership')

    def update(self, items):
        raise ZenpyException("You cannot update Organization Memberships!")


class SatisfactionRatingApi(ModifiableApi):
    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='satisfaction_rating')

    def create(self, ticket_id, satisfaction_rating):
        """
        Create/update a Satisfaction Rating for a ticket.

        :param ticket_id: id of Ticket to rate
        :param satisfaction_rating: SatisfactionRating object.
        """

        self._check_type(satisfaction_rating)
        object_type, payload = self._get_type_and_payload(satisfaction_rating)
        return self._do(self.post,
                        payload=payload,
                        endpoint=Endpoint.satisfaction_ratings.create,
                        endpoint_kwargs=dict(id=ticket_id))


class MacroApi(CRUDApi):
    def __init__(self, subdomain, session, timeout):
        Api.__init__(self, subdomain, session, Endpoint.macros, timeout=timeout, object_type='macro')

    def apply(self, macro_id):
        """
        Show what a macro would do - https://developer.zendesk.com/rest_api/docs/core/macros#show-changes-to-ticket

        :param macro_id: id of macro to test
        """

        return self._get_items(self.endpoint.apply, 'result', id=macro_id)


class TicketApi(RateableApi, TaggableApi, IncrementalApi, CRUDApi):
    """
    The TicketApi adds some Ticket specific functionality
    """

    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='ticket')

    def organizations(self, org_id):
        """
        Retrieve the tickets for this organization.

        :param org_id: organization id
        """
        return self._get_items(self.endpoint.organizations, 'ticket', id=org_id)

    def recent(self):
        """
        Retrieve the most recent tickets
        """
        return self._get_items(self.endpoint.recent, 'ticket', id=None)

    def comments(self, ticket_id):
        """
        Retrieve the comments for a ticket.

        :param ticket_id: ticket id
        """
        return self._get_items(self.endpoint.comments, 'comment', id=ticket_id)

    def events(self, start_time):
        """
        Retrieve TicketEvents
        :param start_time: time to retrieve events from.
        """
        return self._get_items(self.endpoint.events, 'ticket_event', start_time=start_time)

    def audits(self, ticket_id):
        """
        Retrieve TicketAudits.
        :param ticket_id: ticket id
        """
        return self._get_items(self.endpoint.audits, 'ticket_audit', id=ticket_id)

    def metrics(self, ticket_id):
        """
        Retrieve TicketMetric.
        :param ticket_id: ticket id
        """
        return self._get_items(self.endpoint.metrics, 'ticket_metric', id=ticket_id)

    def metrics_incremental(self, start_time):
        """
        Retrieve TicketMetric incremental
        :param start_time: time to retrieve events from.
        """
        return self._get_items(self.endpoint.metrics.incremental, 'ticket_metric_events', start_time=start_time)

    def show_macro_effect(self, ticket, macro):
        """
        Apply macro to ticket. Returns what it *would* do, does not alter the ticket.

        :param ticket: Ticket or ticket id to target
        :param macro_id: Macro or macro id to use
        """

        if isinstance(ticket, Ticket):
            ticket = ticket.id
        if isinstance(macro, Macro):
            macro = macro.id

        response = self._get(
            url=self._get_url(self.endpoint.macro(ticket, macro))
        )
        self._check_and_cache_response(response)
        return self._build_response(response.json())


class TicketImportAPI(CRUDApi):
    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='ticket')

    def __call__(self, *args, **kwargs):
        raise ZenpyException("You must pass ticket objects to this endpoint!")

    def update(self, items):
        raise ZenpyException("You cannot update objects using ticket_import endpoint!")

    def delete(self, items):
        raise ZenpyException("You cannot delete objects using the ticket_import endpoint!")


class RequestAPI(CRUDApi):
    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='request')

    def open(self):
        """
        Return all open requests
        """
        return self._get_items(self.endpoint.open, 'request', sideload=False)

    def solved(self):
        """
        Return all solved requests
        """
        return self._get_items(self.endpoint.solved, 'request', sideload=False)

    def ccd(self):
        """
        Return all ccd requests
        """
        return self._get_items(self.endpoint.ccd, 'request', sideload=False)

    def comments(self, request_id):
        """
        Return comments for request
        """
        return self._get_items(self.endpoint.comments, 'comment', sideload=False, id=request_id)

    def delete(self, items):
        raise ZenpyException("You cannot delete requests!")

    def search(self, *args, **kwargs):
        """
        Search for requests. See the Zendesk docs for more information on the syntax
         https://developer.zendesk.com/rest_api/docs/core/requests#searching-requests
        """
        return self._get_items(self.endpoint.search, 'request', *args, **kwargs)


class SharingAgreementAPI(CRUDApi):
    def __init__(self, subdomain, session, endpoint, timeout):
        Api.__init__(self, subdomain, session, endpoint, timeout=timeout, object_type='sharing_agreement')



from datetime import datetime, timedelta

__author__ = 'facetoe'

import logging

log = logging.getLogger(__name__)


class ResultGenerator(object):
    """
    Generator for handling pagination.
    """

    # TODO fix this
    endpoint_mapping = {
        'user': 'users',
        'ticket': 'tickets',
        'group': 'groups',
        'results': 'results',
        'organization': 'organizations',
        'topic': 'topics',
        'comment': 'comments',
        'ticket_event': 'ticket_events',
        'ticket_audit': 'audits',
        'tag': 'tags',
        'suspended_ticket': 'suspended_tickets',
        'satisfaction_rating': 'satisfaction_ratings',
        'sharing_agreement': 'sharing_agreements',
        'activity': 'activities',
        'group_membership': 'group_memberships',
        'ticket_metric': 'ticket_metrics',
        'ticket_metric_events': 'ticket_metric_events',
        'request': 'requests',
        'user_field': 'user_fields',
        'organization_field': 'organization_fields',
        'brand': 'brands',
        'ticket_field': 'ticket_fields',
        'organization_membership': 'organization_memberships',
        'organization_memberships': 'organization_memberships',
        'macro':  'macros'
    }

    def __init__(self, api, result_key, _json):
        self.api = api
        self._json = _json
        self.result_key = self.endpoint_mapping[result_key]
        self.values = _json[self.result_key]
        self.position = 0
        self.update_attrs(self._json)

    def __iter__(self):
        return self

    def __len__(self):
        return self.count

    def __next__(self):
        return self.next()

    def next(self):
        # Update end_time etc
        self.update_attrs(self._json)

        # Pagination
        if self.position >= len(self.values):
            end_time = self._json.get('end_time', None)
            next_page = self._json.get('next_page', None)

            # If we are calling an incremental API, make sure to honour the restrictions
            if end_time:
                # Incremental APIs return values in blocks of 1000 and the count value is
                # decremented by 1000 for each block that is consumed. If we have < 1000
                # then there is nothing left to retrieve.
                if self._json.get('count', 0) < 1000:
                    raise StopIteration

                # We can't request updates from an incremental api if the
                # start_time value is less than 5 minutes in the future.
                # (end_time is added as start_time to the next_page URL)
                if (datetime.fromtimestamp(int(end_time)) + timedelta(minutes=5)) > datetime.now():
                    raise StopIteration

            if next_page:
                self._json = self._get_as_json(next_page)
                self.values = self._json[self.result_key]
                self.position = 0
            else:
                raise StopIteration()

        if not self.values:
            raise StopIteration()

        item_json = self.values[self.position]
        self.position += 1
        if 'result_type' in item_json:
            object_type = item_json.pop('result_type')
        else:
            # Multiple results have a plural key, however the object_type is singular
            object_type = self.get_singular(self.result_key)
        return self.api.object_manager.object_from_json(object_type, item_json)

    def update_attrs(self, _json):
        # Add attributes such as count/end_time that can be present
        for key, value in _json.items():
            if key != self.result_key:
                setattr(self, key, value)

    def _get_as_json(self, url):
        log.debug("GENERATOR: " + url)
        response = self.api._get(url)
        return response.json()

    def get_singular(self, result_key):
        if result_key.endswith('ies'):
            object_type = result_key.replace('ies', 'y')
        else:
            object_type = result_key[:-1]
        return object_type



import logging
import time
from datetime import datetime

from dateutil.tz import tzutc

from zenpy.lib.exception import ZenpyException
from zenpy.lib.util import is_timezone_aware, is_iterable_but_not_string

__author__ = 'facetoe'

try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    str = str
    unicode = str
    bytes = bytes
    basestring = (str, bytes)
else:
    # 'unicode' exists, must be Python 2
    str = str
    unicode = unicode
    bytes = str
    basestring = basestring

log = logging.getLogger(__name__)


class BaseEndpoint(object):
    """
    BaseEndpoint supplies common formatting operations.
    """

    def __init__(self, endpoint, sideload=None):
        self.endpoint = endpoint
        self.sideload = sideload or []

    @staticmethod
    def _single(endpoint, user_id):
        return "%s/%s%s" % (endpoint, user_id, '.json')

    def _many(self, endpoint, user_ids, action='show_many.json?ids='):
        return "%s/%s%s" % (endpoint, action, self._format_many(user_ids))

    @staticmethod
    def _format(*args, **kwargs):
        return '+'.join(['%s%s' % (key, value) for (key, value) in kwargs.items()] + [a for a in args])

    @staticmethod
    def _format_many(items):
        return ",".join([str(i) for i in items])

    def _format_sideload(self, items, seperator='&'):
        if isinstance(items, basestring):
            items = [items]
        return '%sinclude=%s' % (seperator, self._format_many(items))


class PrimaryEndpoint(BaseEndpoint):
    """
    A PrimaryEndpoint takes an id or list of ids and either returns the objects
    associated with them or performs actions on them (eg, update/delete).
    """

    ISO_8601_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

    def __call__(self, **kwargs):
        query = ""
        modifiers = []
        for key, value in kwargs.items():
            if key == 'id':
                query += self._single(self.endpoint, value)
            elif key == 'ids':
                query += self._many(self.endpoint, value)
            elif key == 'destroy_ids':
                query += self._many(self.endpoint, value, action='destroy_many.json?ids=')
            elif key == 'create_many':
                query = "".join([self.endpoint, '/create_many.json'])
            elif key == 'create_or_update_many':
                query = self.endpoint
            elif key == 'recover_ids':
                query = self._many(self.endpoint, value, action='recover_many.json?ids=')
            elif key == 'update_many':
                query = "".join([self.endpoint, '/update_many.json'])
            elif key in ('sort_by', 'sort_order'):
                modifiers.append((key, value))
            elif key == 'since':
                modifiers.append((key, value.strftime(self.ISO_8601_FORMAT)))

        if modifiers:
            query += '&' + "&".join(["%s=%s" % (k, v) for k, v in modifiers])

        if self.endpoint not in query:
            query = self.endpoint + '.json?' + query

        if 'sideload' in kwargs and not kwargs['sideload']:
            return query
        else:
            return query + self._format_sideload(self.sideload)


class SecondaryEndpoint(BaseEndpoint):
    """
    A SecondaryEndpoint takes a single ID and returns the
    object associated with it.
    """

    def __call__(self, **kwargs):
        if not kwargs:
            raise ZenpyException("This endpoint requires arguments!")
        return self.endpoint % kwargs


class MutlipleIDEndpoint(BaseEndpoint):
    def __call__(self, *args):
        if not args or len(args) < 2:
            raise ZenpyException("This endpoint requires at least two arguments!")
        return self.endpoint.format(*args)


class IncrementalEndpoint(BaseEndpoint):
    """
    An IncrementalEndpoint takes a start_time parameter
    for querying the incremental api endpoint.

    Note: The Zendesk API expects UTC time. If a timezone aware datetime object is passed
    Zenpy will convert it to UTC, however if a naive object or unix timestamp is passed there is nothing
    Zenpy can do. It is recommended to always pass timezone aware objects to this endpoint.

    :param start_time: Unix timestamp or datetime object
    """

    def __call__(self, start_time=None):
        if not start_time:
            raise ZenpyException("Incremental Endoint requires a start_time parameter!")

        if isinstance(start_time, datetime):
            if is_timezone_aware(start_time):
                start_time = start_time.astimezone(tzutc())
            else:
                log.warning(
                    "Non timezone-aware datetime object passed to IncrementalEndpoint. "
                    "The Zendesk API expects UTC time, if this is not the case results will be incorrect!"
                )
            unix_time = time.mktime(start_time.timetuple())
        else:
            unix_time = start_time
        query = "start_time=%s" % str(unix_time)
        return self.endpoint + query + self._format_sideload(self.sideload, seperator='&')


class AttachmentEndpoint(BaseEndpoint):
    def __call__(self, **kwargs):
        query = self.endpoint
        for key, value in kwargs.items():
            if value:
                if not '&' in query:
                    query += '&'
                query += '{}={}'.format(key, value)
        return query


class SearchEndpoint(BaseEndpoint):
    """
    The search endpoint accepts all the parameters defined in the Zendesk `Search Documentation <https://developer.zendesk.com/rest_api/docs/core/search>`_.
    Zenpy defines several keywords that are mapped to the Zendesk comparison operators:

    +-----------------+------------------+
    | **Keyword**     | **Operator**     |
    +-----------------+------------------+
    | keyword         | : (equality)     |
    +-----------------+------------------+
    | \*_greater_than | > (numeric|type) |
    +-----------------+------------------+
    | \*_less_than    | < (numeric|type) |
    +-----------------+------------------+
    | \*_after        | > (time|date)    |
    +-----------------+------------------+
    | \*_before       | < (time|date)    |
    +-----------------+------------------+
    | minus           | \- (negation)    |
    +-----------------+------------------+
    | \*_between      | > < (dates only) |
    +-----------------+------------------+

    For example the call:

    .. code:: python

      zenpy.search("zenpy", created_between=[yesterday, today], type='ticket', minus='negated')

    Would generate the following API call:
    ::
        /api/v2/search.json?query=zenpy+created>2015-08-29 created<2015-08-30+type:ticket+-negated


    """

    ZENDESK_DATE_FORMAT = "%Y-%m-%d"

    def __call__(self, *args, **kwargs):

        renamed_kwargs = dict()
        modifiers = list()
        sort_order = list()
        for key, value in kwargs.items():
            if isinstance(value, datetime):
                kwargs[key] = value.strftime(self.ZENDESK_DATE_FORMAT)
            elif is_iterable_but_not_string(value) and key == 'ids':
                kwargs[key] = self._format_many(value)

            if key.endswith('_between'):
                modifiers.append(self.format_between(key, value))
            elif key in ('sort_by', 'sort_order'):
                sort_order.append("%s=%s" % (key, value))
            elif key.endswith('_after'):
                renamed_kwargs[key.replace('_after', '>')] = kwargs[key]
            elif key.endswith('_before'):
                renamed_kwargs[key.replace('_before', '<')] = kwargs[key]
            elif key.endswith('_greater_than'):
                renamed_kwargs[key.replace('_greater_than', '>')] = kwargs[key]
            elif key.endswith('_less_than'):
                renamed_kwargs[key.replace('_less_than', '<')] = kwargs[key]
            elif key == 'minus':
                if is_iterable_but_not_string(value):
                    [modifiers.append("-%s" % v) for v in value]
                else:
                    modifiers.append("-%s" % value)
            elif is_iterable_but_not_string(value):
                modifiers.append(self.format_or(key, value))
            else:
                renamed_kwargs.update({key + ':': '"%s"' % value})

        query = self.endpoint + 'query='
        if args:
            query += ' '.join(args) + '+'

        sort_section = ""
        if sort_order:
            sort_section += '&' + "&".join(sort_order)

        search_parameters = self._format(*modifiers, **renamed_kwargs)

        return "%(query)s%(search_parameters)s%(sort_section)s" % locals()

    def format_between(self, key, values):
        if not is_iterable_but_not_string(values):
            raise ZenpyException("*_between requires an iterable (list, set, tuple etc)")
        elif not len(values) == 2:
            raise ZenpyException("*_between requires exactly 2 items!")
        elif not all([isinstance(d, datetime) for d in values]):
            raise ZenpyException("*_between only works with dates!")
        key = key.replace('_between', '')
        dates = [v.strftime(self.ZENDESK_DATE_FORMAT) for v in values]
        return "%s>%s %s<%s" % (key, dates[0], key, dates[1])

    def format_or(self, key, values):
        return " ".join(['%s:"%s"' % (key, v) for v in values])


class RequestSearchEndpoint(BaseEndpoint):
    def __call__(self, *args, **kwargs):
        if not args:
            raise ZenpyException("You need to pass the query string as the first parameter")

        query = "query=%s" % args[0]
        result = []
        for key, value in kwargs.items():
            result.append("%s=%s" % (key, value))
        query += '&' + "&".join(result)
        return self.endpoint + query


class SatisfactionRatingEndpoint(BaseEndpoint):
    def __call__(self, score=None, sort_order=None):
        if sort_order not in ('asc', 'desc'):
            raise ZenpyException("sort_order must be one of (asc, desc)")

        base_url = self.endpoint + '?'
        if score:
            result = base_url + "score={}".format(score)
        else:
            result = base_url

        if sort_order:
            result += '&sort_order={}'.format(sort_order)
        return result


class MacroEndpoint(BaseEndpoint):
    def __call__(self, sort_order=None, sort_by=None, **kwargs):
        kwargs.pop('sideload', None)
        if sort_order and sort_order not in ('asc', 'desc'):
            raise ZenpyException("sort_order must be one of (asc, desc)")
        if sort_by and sort_by not in ('alphabetical', 'created_at', 'updated_at', 'usage_1h', 'usage_24h', 'usage_7d'):
            raise ZenpyException(
                "sort_by is invalid - https://developer.zendesk.com/rest_api/docs/core/macros#available-parameters")

        if 'id' in kwargs:
            if len(kwargs) > 1:
                raise ZenpyException("When specifying an id it must be the only parameter")
            url_out = ''
        else:
            url_out = self.endpoint + '?'

        for key, value in kwargs.items():
            if isinstance(value, bool):
                value = str(value).lower()
            if key == 'id':
                url_out += self._single(self.endpoint, value)
            else:
                url_out += '&{}={}'.format(key, value)

        if sort_order:
            url_out += '&sort_order={}'.format(sort_order)
        if sort_by:
            url_out += '&sort_by={}'.format(sort_by)
        return url_out


class Endpoint(object):
    """
    The Endpoint object ties it all together.
    """

    activities = PrimaryEndpoint('activities')
    attachments = PrimaryEndpoint('attachments')
    attachments.upload = AttachmentEndpoint('uploads.json?')
    brands = PrimaryEndpoint('brands')
    end_user = SecondaryEndpoint('end_users/%(id)s.json')
    group_memberships = PrimaryEndpoint('group_memberships')
    groups = PrimaryEndpoint('groups', ['users'])
    job_statuses = PrimaryEndpoint('job_statuses')
    macros = MacroEndpoint('macros')
    macros.apply = SecondaryEndpoint('macros/%(id)s/apply.json')
    organization_memberships = PrimaryEndpoint('organization_memberships')
    organizations = PrimaryEndpoint('organizations')
    organizations.external = SecondaryEndpoint('organizations/search.json?external_id=%(id)s')
    organizations.incremental = IncrementalEndpoint('incremental/organizations.json?')
    organizations.organization_fields = PrimaryEndpoint('organization_fields')
    organizations.organization_memberships = SecondaryEndpoint('organizations/%(id)s/organization_memberships.json')
    organizations.requests = SecondaryEndpoint('organizations/%(id)s/requests.json')
    organizations.tags = SecondaryEndpoint('organizations/%(id)s/tags.json')
    requests = PrimaryEndpoint('requests')
    requests.ccd = PrimaryEndpoint("requests/ccd")
    requests.comments = SecondaryEndpoint('requests/%(id)s/comments.json')
    requests.open = PrimaryEndpoint("requests/open")
    requests.search = RequestSearchEndpoint('requests/search.json?')
    requests.solved = PrimaryEndpoint("requests/solved")
    satisfaction_ratings = SatisfactionRatingEndpoint('satisfaction_ratings')
    satisfaction_ratings.create = SecondaryEndpoint('tickets/%(id)s/satisfaction_rating.json')
    search = SearchEndpoint('search.json?')
    sharing_agreements = PrimaryEndpoint('sharing_agreements')
    suspended_tickets = PrimaryEndpoint('suspended_tickets')
    suspended_tickets.recover = SecondaryEndpoint('suspended_tickets/%(id)s/recover.json')
    tags = PrimaryEndpoint('tags')
    ticket_fields = PrimaryEndpoint('ticket_fields')
    ticket_import = PrimaryEndpoint('imports/tickets')
    ticket_metrics = PrimaryEndpoint('ticket_metrics')
    tickets = PrimaryEndpoint('tickets', ['users', 'groups', 'organizations'])
    tickets.audits = SecondaryEndpoint('tickets/%(id)s/audits.json')
    tickets.comments = SecondaryEndpoint('tickets/%(id)s/comments.json')
    tickets.events = IncrementalEndpoint('incremental/ticket_events.json?')
    tickets.incremental = IncrementalEndpoint('incremental/tickets.json?',
                                              sideload=['users', 'groups', 'organizations'])
    tickets.metrics = SecondaryEndpoint('tickets/%(id)s/metrics.json')
    tickets.metrics.incremental = IncrementalEndpoint('incremental/ticket_metric_events.json?')
    tickets.organizations = SecondaryEndpoint('organizations/%(id)s/tickets.json')
    tickets.recent = SecondaryEndpoint('tickets/recent.json')
    tickets.tags = SecondaryEndpoint('tickets/%(id)s/tags.json')
    tickets.macro = MutlipleIDEndpoint('tickets/{0}/macros/{1}/apply.json')
    topics = PrimaryEndpoint('topics')
    topics.tags = SecondaryEndpoint('topics/%(id)s/tags.json')
    user_fields = PrimaryEndpoint('user_fields')
    users = PrimaryEndpoint('users', ['organizations', 'abilities', 'roles', 'identities', 'groups'])
    users.assigned = SecondaryEndpoint('users/%(id)s/tickets/assigned.json')
    users.cced = SecondaryEndpoint('users/%(id)s/tickets/ccd.json')
    users.create_or_update = PrimaryEndpoint('users/create_or_update')
    users.create_or_update_many = PrimaryEndpoint('users/create_or_update_many.json')
    users.group_memberships = SecondaryEndpoint('users/%(id)s/group_memberships.json')
    users.groups = SecondaryEndpoint('users/%(id)s/groups.json')
    users.incremental = IncrementalEndpoint('incremental/users.json?')
    users.me = SecondaryEndpoint('users/me.json')
    users.merge = SecondaryEndpoint('users/%(id)s/merge.json')
    users.organization_memberships = SecondaryEndpoint('users/%(id)s/organization_memberships.json')
    users.organizations = SecondaryEndpoint('users/%(id)s/organizations.json')
    users.related = SecondaryEndpoint('users/%(id)s/related.json')
    users.requested = SecondaryEndpoint('users/%(id)s/tickets/requested.json')
    users.requests = SecondaryEndpoint('users/%(id)s/requests.json')
    users.tags = SecondaryEndpoint('users/%(id)s/tags.json')



import cachetools

from zenpy.lib.exception import ZenpyCacheException

__author__ = 'facetoe'


class ZenpyCache(object):
    """
    Wrapper class for the various cachetools caches. Adds ability to change cache implementations
    on the fly and change the maxsize setting.
    """
    AVAILABLE_CACHES = [c for c in dir(cachetools) if c.endswith('Cache') and c != 'Cache']

    def __init__(self, cache_impl, maxsize, **kwargs):
        self.cache = self._get_cache_impl(cache_impl, maxsize, **kwargs)

    def set_cache_impl(self, cache_impl, maxsize, **kwargs):
        """
        Change cache implementation. The contents of the old cache will
        be transferred to the new one.
        :param cache_impl: Name of cache implementation, must exist in AVAILABLE_CACHES
        """
        new_cache = self._get_cache_impl(cache_impl, maxsize, **kwargs)
        self._populate_new_cache(new_cache)
        self.cache = new_cache

    def pop(self, key, default=None):
        return self.cache.pop(key, default)

    def items(self):
        return self.cache.items()

    @property
    def impl_name(self):
        """
        Name of current cache implementation
        """
        return self.cache.__class__.__name__

    @property
    def maxsize(self):
        """
        Current max size
        """
        return self.cache.maxsize

    def set_maxsize(self, maxsize, **kwargs):
        """
        Set maxsize. This involves creating a new cache and transferring the items.
        """
        new_cache = self._get_cache_impl(self.impl_name, maxsize, **kwargs)
        self._populate_new_cache(new_cache)
        self.cache = new_cache

    @property
    def currsize(self):
        return len(self.cache)

    def _populate_new_cache(self, new_cache):
        for key, value in self.cache.items():
            new_cache[key] = value

    def _get_cache_impl(self, cache_impl, maxsize, **kwargs):
        if cache_impl not in self.AVAILABLE_CACHES:
            raise ZenpyCacheException(
                    "No such cache: %s, available caches: %s" % (cache_impl, str(self.AVAILABLE_CACHES)))
        return getattr(cachetools, cache_impl)(maxsize, **kwargs)

    def __iter__(self):
        return self.cache.__iter__()

    def __getitem__(self, item):
        return self.cache[item]

    def __setitem__(self, key, value):
        self.cache[key] = value

    def __delitem__(self, key):
        del self.cache[key]

    def __contains__(self, item):
        return item in self.cache



__author__ = 'facetoe'


class ZenpyException(Exception):
    """
    A ZenpyException is raised for internal errors
    """


class APIException(Exception):
    """
    An APIException is raised when the API rejects a query.
    """


class RecordNotFoundException(Exception):
    """
    A RecordNotFoundException is raised when the API cannot find a record
    """


class ZenpyCacheException(Exception):
    """
    A ZenpyCacheException is raised for errors relating the the ZenpyCache
    """
